from os.path import join
from collections.abc import Iterable
from typing import Dict, List, Callable
import fsspec
import yaml
import warnings

import numpy as np
import xarray as xr
from zarr import NestedDirectoryStore

from datetime import datetime, timedelta
from cftime import DatetimeJulian

from .utils import batched

class UFSDataset:
    """Open and store a UFS generated NetCDF dataset to zarr from a single model component (FV3, MOM6, CICE6). The two main methods that are useful are :meth:`open_dataset` and :meth:`store_dataset`.

    Note:
        The ``path_in`` argument on __init__ probably needs some attention, especially in relation to the ``file_prefixes`` option. This should be addressed once we start thinking about datasets other than replay.

    Required Fields in Config:
        path_out (str): the outermost directory to store the dataset
        forecast_hours (list of int): with the forecast hours to save
        file_prefixes (list of str): with the filename prefixes to read, e.g. ["sfg", "bfg"] to read all files starting with this prefix at this cycle, for all forecast_hours

    Optional Fields in Config:
        coords (list of str): containing static coordinate variables to store only one time
        coords_path_out (str): path to store coordinates, if not provided then store under 'coordinates' directory in data_path
        data_vars (list of str): containing variables that evolve in time to be stored if not provided, all variables will be stored
        chunks_in, chunks_out (dict): containing chunksizes for each dimension

    Args:
        path_in (callable): map the following arguments to a path (str):
        config_filename (str): to yaml file containing the overall configuration
        is_nested (bool, optional): if True write a :class:`NestedDirectoryStore`, otherwise write Zarr's default directory structure

    Sets Attributes:
        path_out (str): the outermost directory to store the dataset
        forecast_hours (list of int): with the forecast hours to save
        file_prefixes (list of str): with the filename prefixes inside of each cycle's directory,
        coords, data_vars (list): with variable names of coordinates and data variables
        chunks_in, chunks_out (dict): specifying how to chunk the data when reading and writing
        config (dict): with the configuration provided by the file
        is_nested (bool): whether or not to write with :class:`NestedDirectoryStore`
    """

    @property
    def data_path(self):
        """Where to write forecast data variables to"""
        return self._join(self.path_out, self.zarr_name)

    @property
    def coords_path(self) -> str:
        """Where to write static coordinates to"""
        if self.coords_path_out is None:
            return self._join(self.path_out, "coordinates", self.zarr_name)
        else:
            return self._join(self.coords_path_out, self.zarr_name)

    @property
    def default_open_dataset_kwargs(self) -> Dict:
        kw = {
            "parallel": True,
            "chunks": self.chunks_in,
            "decode_times": True,
            "preprocess": self._preprocess,
        }
        return kw

    def __init__(self, path_in: Callable, config_filename: str, is_nested: bool = False) -> None:
        super(UFSDataset, self).__init__()
        name = self.__class__.__name__  # e.g., FV3Dataset, MOMDataset

        # create and initialze instance variable for class attributes
        self.path_in: Callable = path_in
        self.is_nested: bool = is_nested
        self.path_out: str = ""
        self.forecast_hours: List[int] = []
        self.file_prefixes: List[str] = []

        self.chunks_in: Dict = {}
        self.chunks_out: Dict = {}
        self.coords: List[str] = []
        self.data_vars: List[str] = []
        self.zarr_name: str = ""
        self.coords_path_out = None

        with open(config_filename, "r") as f:
            contents = yaml.safe_load(f)
            self.config = contents[name]

        # look for these requited inputs
        for key in ["path_out", "forecast_hours", "file_prefixes"]:
            try:
                setattr(self, key, self.config[key])
            except KeyError:
                raise KeyError(f"{name}.__init__: Could not find {key} in {config_filename}, but this is required")

        # look for these optional inputs
        for key in ["chunks_in", "chunks_out", "coords", "data_vars", "coords_path_out"]:
            if key in self.config:
                setattr(self, key, self.config[key])
            else:
                print(f"{name}.__init__: Could not find {key} in {config_filename}, using default.")

        # warn user about not finding coords
        if len(self.coords) == 0:
            warnings.warn(
                f"{name}.__init__: Could not find 'coords' in {config_filename}, will not store coordinate data"
            )

        if len(self.data_vars) == 0:
            warnings.warn(
                f"{name}.__init__: Could not find 'data_vars' in {config_filename}, will store all data variables"
            )

        # check that file_prefixes is a list
        self.file_prefixes = [self.file_prefixes] if isinstance(self.file_prefixes, str) else self.file_prefixes

    def open_dataset(self, cycles: datetime, fsspec_kwargs=None, **kwargs):
        """Read data from specified DA cycles

        Args:
            cycles (datetime.datetime or List[datetime.datetime]): datetime object(s) giving initial time for each DA cycles
            fsspec_kwargs (dict, optional): optional arguments passed to :func:`fsspec.open_files`
            **kwargs (dict, optional): optional arguments passed to :func:`xarray.open_mfdataset`, in addition to
                the ones provided by :attr:`default_open_dataset_kwargs`

        Returns:
            xds (xarray.Dataset): with output from a single model component (e.g., FV3)
                and a single forecast window, with all output times in
                that window merged into one dataset
        """

        kw = self.default_open_dataset_kwargs.copy()
        kw.update(kwargs)

        fnames = self.path_in(cycles, self.forecast_hours, self.file_prefixes)

        # Maybe there's a more elegant way to handle this, but with local files, fsspec closes them
        # before dask reads them...
        if fsspec_kwargs is None:
            xds = xr.open_mfdataset(fnames, **kw)
        else:
            with fsspec.open_files(fnames, **fsspec_kwargs) as files:
                # This assumes netcdf4, which I think is a safe fallback for UFS data
                engine = kw.pop("engine", "netcdf4")
                if engine == "netcdf4":
                    xds = xr.open_mfdataset([this_file.name for this_file in files], **kw)
                else:
                    xds = xr.open_mfdataset(files, **kw)
        return xds

    def chunk(self, xds):
        """Using the yaml-provided or default chunking scheme, chunk all arrays in this dataset

        Note:
            This should probably be replaced with rechunker https://rechunker.readthedocs.io/en/latest/

        Args:
            xds (xarray.Dataset): as provided by :meth:`open_dataset`

        Returns:
            xds (xarray.Dataset): rechunked as specified
        """

        chunks = self.chunks_out.copy()
        for key in self.chunks_out.keys():
            if key not in xds.dims:
                chunks.pop(key)

        xds = xds.transpose(*list(chunks.keys()))
        return xds.chunk(chunks)

    def store_dataset(self,
        xds: xr.Dataset,
        store_coords: bool = False,
        coords_kwargs=None,
        **kwargs) -> None:
        """Open all netcdf files for this model component and at this DA window, store
        coordinates one time only, select data based on
        desired forecast hour, then store it.

        Args:
            xds (xarray.Dataset): as provided by :meth:`open_dataset`
            store_coords (bool, optional): if True, store coordinates in separate directory
            coords_kwargs (dict, optional): passed to :func:`xarray.to_zarr` via :meth:`_store_coordinates`
            kwargs (dict): optional arguments passed to :func:`xarray.to_zarr` via :meth:`_store_data_vars`
        """

        xds = xds.reset_coords()

        # need to store coordinates dataset only one time
        if store_coords:
            coords = [x for x in self.coords if x in xds]
            cds = xds[coords].set_coords(coords)
            if "member" in cds:
                cds = cds.isel(member=0).drop("member")
            ckw = dict() if coords_kwargs is None else coords_kwargs
            self._store_coordinates(cds, **ckw)

        # now data variables at this cycle
        # make various time variables as coordinates
        xds = xds.set_coords(["time", "cftime", "ftime"])
        if len(self.data_vars) > 0:
            data_vars = [x for x in self.data_vars if x in xds]
            xds = xds[data_vars]

        self._store_data_vars(xds, **kwargs)

    def _store_coordinates(self, cds: xr.Dataset, **kwargs) -> None:
        """Store the static coordinate information to zarr

        Args:
            cds (xarray.Dataset): with only the static coordinate information
        """

        try:
            assert len(cds.data_vars) == 0
        except AssertionError:
            msg = "UFSDataset._store_coordinates: "+\
                f"We should not have any data variables in this dataset, but we found some."+\
                f"\n{cds.data_vars}"
            raise AttributeError(msg)

        # these don't need to be chunked, coordinates are opened in memory
        store = NestedDirectoryStore(path=self.coords_path) if self.is_nested else self.coords_path
        cds.to_zarr(store, **kwargs)
        print(f"Stored coordinate dataset at {self.coords_path}")

    def _store_data_vars(self, xds: xr.Dataset, **kwargs) -> None:
        """Store the data variables

        Args:
            xds (xarray.Dataset): the big dataset with all desired data variables, for this model component
                and at this particular DA window
        """

        xds = self.chunk(xds)

        store = NestedDirectoryStore(path=self.data_path) if self.is_nested else self.data_path
        xds.to_zarr(store, **kwargs)
        print(f"Stored dataset at {self.data_path}")

    @staticmethod
    def _preprocess(xds):
        """Used to remove a redundant surface pressure found in both physics and dynamics FV3 files,
        which are slightly different and so cause a conflict. This method is not used on its own, but
        given as an option to :func:`xarray.open_mfdataset`

        Args:
            xds (xarray.Dataset): A single netcdf file from the background forecast

        Returns:
            xds (xarray.Dataset): with ``pressfc`` variable removed if this is from FV3 dynamics output
        """
        # We can't rely on xds.encoding when reading from s3, so have to infer if this is dynamics
        # vs physics dataset by the other fields that exist in the dataset
        dyn_vars = ["tmp", "ugrd", "vgrd", "spfh", "o3mr"]
        if "pressfc" in xds.data_vars and any(v in xds.data_vars for v in dyn_vars):
            del xds["pressfc"]
        return xds

    @staticmethod
    def _cftime2time(cftime):
        """Convert cftime array to numpy.datetime64

        Args:
            cftime (xarray.DataArray): with DatetimeJulian objects

        Returns:
            xtime (xarray.DataArray): with numpy.datetime64 objects
        """
        time = np.array(
            [
                np.datetime64(
                    datetime(
                        int(t.dt.year),
                        int(t.dt.month),
                        int(t.dt.day),
                        int(t.dt.hour),
                        int(t.dt.minute),
                        int(t.dt.second),
                    )
                )
                for t in cftime
            ]
        )
        xtime = xr.DataArray(
            time,
            coords=cftime.coords,
            dims=cftime.dims,
            attrs={
                "long_name": "time",
                "axis": "T",
            },
        )
        return xtime

    @staticmethod
    def _time2cftime(time):
        """Convert numpy.datetime64 array to cftime

        Args:
            time (xarray.DataArray): with numpy.datetime64 objects

        Returns:
            xcftime (xarray.DataArray): with DatetimeJulian objects
        """
        cftime = np.array(
            [
                DatetimeJulian(
                    int(t.dt.year),
                    int(t.dt.month),
                    int(t.dt.day),
                    int(t.dt.hour),
                    int(t.dt.minute),
                    int(t.dt.second),
                    has_year_zero=False,
                )
                for t in time
            ]
        )
        xcftime = xr.DataArray(
            cftime,
            coords=time.coords,
            dims=time.dims,
            attrs={
                "long_name": "time",
                "axis": "T",
                "calendar_type": "JULIAN",
            },
        )
        return xcftime

    @staticmethod
    def _time2ftime(time, cycles):
        """Compute the ftime (forecast_time) array, indicating the hours since
        initialization for each timestamp

        Args:
            time (xr.DataArray): with numpy.datetime64 objects
            cycles (datetime or List[datetime]): with the DA cycle(s) to grab

        Returns:
            xftime (xr.DataArray): forecast_time
        """
        cycles = [cycles] if not isinstance(cycles, Iterable) else cycles
        n_output_per_cycle = len(time) // len(cycles)
        time_batches = list(batched(time.values, n_output_per_cycle))
        ftime = np.array([
            these_times - np.datetime64(this_cycle) for these_times, this_cycle in zip(time_batches, cycles)
        ]).flatten()
        xftime = xr.DataArray(
            ftime,
            coords=time.coords,
            dims=time.dims,
            attrs={
                "long_name": "forecast_time",
                "description": f"time passed since forecast initialization",
                "axis": "T",
            },
        )
        return xftime

    @staticmethod
    def _join(a, *p):
        """System independent join operation"""
        clouds = ("gcs://", "s3://", "https://")
        if any(x in a for x in clouds) or any(any(x in this_path for x in clouds) for this_path in p):
            try:
                assert isinstance(a, str) and all(isinstance(this_path, str) for this_path in p)
            except:
                raise TypeError(f"For cloud storage, paths need to be strings.")

            path = a
            join_char = "/" if a[-1] != "/" else ""
            for this_path in p:
                path += join_char
                path += this_path
                join_char = "/" if this_path[-1] != "/" else ""

            return path

        else:
            return join(a, *p)



from os.path import join, isdir
from datetime import datetime, timedelta
import logging
import yaml
from shutil import rmtree
import itertools
from collections.abc import Iterable

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as darray
from zarr import NestedDirectoryStore

from ufs2arco import (
    FV3Dataset, 
    MOM6Dataset, 
    CICE6Dataset, 
    Timer, 
    MOM6Regridder,
    CICE6Regridder
)

class ReplayMover1Degree():

    n_jobs = None

    forecast_hours = None
    file_prefixes = None

    @property
    def xcycles(self):
        return xr.DataArray(self.cycles, coords={"cycles": self.cycles}, dims="cycles")


    @property
    def xtime(self):
        """These are the time stamps of the resulting dataset.

        This was created by processing a few DA cycles with the desired forecast hours, and figuring out
        operations were needed to map from a list of all DA cycles to the resulting ``"time"``
        array in the final dataset.
        """
        return xr.DataArray(self.time, coords={"time": self.time}, dims="time", attrs={"long_name": "time", "axis": "T"})


    @property
    def splits(self):
        """The indices used to split all cycles across :attr:`n_jobs`"""
        return [int(x) for x in np.linspace(0, len(self.xcycles), self.n_jobs+1)]

    def cache_storage(self, job_id):
        """Location to store the s3 data, before subsetting, rechunking, and pushing to GCS

        Args:
            job_id (int): the slurm job id, determines cache storage location

        Returns:
            cache_storage (str): location to store s3 netcdf files
        """
        return join(self.main_cache_path, str(job_id))

    def ods_kwargs(self, job_id):
        """These are passed to xarray.open_dataset to read from s3 and store the file in cache

        Args:
            job_id (int): the slurm job id, determines cache storage location

        Returns:
            xarray_open_dataset_kwargs (dict): job_id determines cache_storage location
        """
        okw = {
            "fsspec_kwargs": {
                "s3": {"anon": True},
                "filecache": {"cache_storage": self.cache_storage(job_id)},
            },
            "engine":"h5netcdf"
        }
        return okw


    def my_cycles(self, job_id):
        """The cycle timestamps for this job

        Args:
            job_id (int): the slurm job id, determines cache storage location

        Returns:
            cycles_datetime (List[datetime]): with the cycle numbers to be processed by this slurm job
        """
        slices = [slice(st, ed) for st, ed in zip(self.splits[:-1], self.splits[1:])]
        xda = self.xcycles.isel(cycles=slices[job_id])
        cycles_datetime = self.npdate2datetime(xda)
        return cycles_datetime


    def __init__(
        self,
        n_jobs,
        config_filename,
        component="fv3",
        storage_options=None,
        main_cache_path=f"/contrib/Tim.Smith/tmp-replay/1.00-degree",
    ):
        self.n_jobs = n_jobs
        self.config_filename = config_filename
        self.component = component
        self.storage_options = storage_options if storage_options is not None else dict()
        self.main_cache_path = main_cache_path

        with open(config_filename, "r") as f:
            config = yaml.safe_load(f)

        name = f"{component.upper()}Dataset" # i.e., FV3Dataset ... maybe an unnecessary generalization at this stage
        self.forecast_hours = config[name]["forecast_hours"]
        self.file_prefixes = config[name]["file_prefixes"]
        self.cycles = pd.date_range(**config[name]["cycles"])
        self.time = pd.date_range(**config[name]["time"])

        self.Regridder = None
        if component.lower() == "fv3":
            self.Dataset = FV3Dataset
            self.cached_path = self.fv3_path
            if "regrid" in self.config.keys():
                raise NotImplementedError
        elif component.lower() == "mom6":
            self.Dataset = MOM6Dataset
            self.cached_path = self.mom6_path
            self.Regridder = MOM6Regridder
        elif component.lower() == "cice6":
            self.Dataset = CICE6Dataset
            self.cached_path = self.cice6_path
            self.Regridder = CICE6Regridder

        # for move_single_dataset, we have to figure out how many resulting timestamps we have
        # within a single DA cycle
        try:
            assert "freq" in config[name]["cycles"].keys()
            assert "freq" in config[name]["time"].keys()

        except:
            raise KeyError("ReplayMover.__init__: we need 'freq' inside the config 'cycles' and 'time' sections")
        delta_t_cycle = pd.Timedelta(config[name]["cycles"]["freq"])
        delta_t_time  = pd.Timedelta(config[name]["time"]["freq"])
        self.n_steps_per_cycle = delta_t_cycle // delta_t_time


    def run(self, job_id):
        """This pulls data and stores it to the desired storage location.
        This is essentially a function that can run completely independently of other objects

        Note:
            This expects :meth:`.store_container` to have been run first

        Args:
            job_id (int): the slurm job id, determines cache storage location
        """

        walltime = Timer()
        localtime = Timer()
        walltime.start()

        store_coords = False # we don't need a separate coords dataset for FV3
        for cycle in self.my_cycles(job_id):

            localtime.start(f"Reading {str(cycle)}")

            try:
                self.move_single_dataset(job_id, cycle)
            except Exception as e:
                logging.exception(e)
                print(f"ReplayMover.run({job_id}): Failed to store {str(cycle)}")
                pass
            localtime.stop()

        walltime.stop("Total Walltime")
    
    def regrid_to_gaussian(
            self,
            xds: xr.Dataset,
            gaussian_grid_path: str,
    ) -> xr.Dataset:
        """Regrid (if needed) - will be used of ocean/sea ice data"""
        gaussian_grid = xr.open_zarr(
            gaussian_grid_path,
            storage_options={"token": "anon"},
            )
        lons = gaussian_grid['grid_xt'].values
        lats = gaussian_grid['grid_yt'].values
        rg = self.Regridder(
            lats1d_out=lats, 
            lons1d_out=lons,
            ds_in=xds, 
            config_filename=self.config_filename,
            )
        xds = rg.regrid(xds)
        
        return xds
        

    def move_single_dataset(self, job_id, cycle):
        """Store a single cycle to zarr"""

        replay = self.Dataset(path_in=self.cached_path, config_filename=self.config_filename)
        xds = replay.open_dataset(cycle, **self.ods_kwargs(job_id))

        if "regrid" in replay.config.keys():
            xds = self.regrid_to_gaussian(
                xds=xds,
                gaussian_grid_path=replay.config['gaussian_grid'],
                )

        index = list(self.xtime.values).index(xds.time.values[0])
        tslice = slice(index, index+self.n_steps_per_cycle)

        # subset the variables here in order to remove extraneous dimensions
        if len(replay.data_vars)>0:
            data_vars = [x for x in replay.data_vars if x in xds]
            xds = xds[data_vars]

        spatial_region = {k: slice(None, None) for k in xds.dims if k != "time"}
        region = {
            "time": tslice,
            **spatial_region,
        }
        region = {k : v for k,v in region.items() if k in xds.dims}

        replay.store_dataset(
            xds,
            region=region,
            storage_options=self.storage_options,
        )
        # This is a hacky way to clear the cache, since we don't create a filesystem object
        del xds
        if isdir(self.cache_storage(job_id)):
            rmtree(self.cache_storage(job_id), ignore_errors=True)


    def store_container(self):
        """Create an empty container that has the write shape, chunks, and dtype for each variable
        """

        localtime = Timer()

        replay = self.Dataset(path_in=self.cached_path, config_filename=self.config_filename)

        localtime.start("Reading Single Dataset")
        cycle = self.my_cycles(0)[0]
        xds = replay.open_dataset(cycle, **self.ods_kwargs(0))

        if "regrid" in replay.config.keys():
            xds = self.regrid_to_gaussian(
                xds=xds,
                gaussian_grid_path=replay.config['gaussian_grid'],
                )

        xds = xds.reset_coords()
        localtime.stop()

        xds = xds.drop(["ftime", "cftime"])
        data_vars = [x for x in replay.data_vars if x in xds]
        xds = xds[data_vars]

        # Make a container, starting with coordinates
        single = self.remove_time(xds)
        dds = xr.Dataset()
        for key in single.coords:
            dds[key] = xds[key]

        localtime.start("Making container for the dataset")
        dds["time"] = self.xtime
        dds = self.add_time_coords(dds, replay._time2cftime)
        for key in single.data_vars:

            dims = ("time",) + single[key].dims
            chunks = tuple(replay.chunks_out[k] for k in dims)
            shape = (len(dds["time"]),) + single[key].shape

            dds[key] = xr.DataArray(
                data=darray.zeros(
                    shape=shape,
                    chunks=chunks,
                    dtype=single[key].dtype,
                ),
                coords={"time": dds["time"], **{d: single[d] for d in single[key].dims}},
                dims=dims,
                attrs=single[key].attrs.copy(),
            )
            print(f"\t ... done with {key}")

        localtime.stop()

        localtime.start("Storing to zarr")
        store = NestedDirectoryStore(path=replay.data_path) if replay.is_nested else replay.data_path
        dds.to_zarr(store, compute=False, storage_options=self.storage_options)
        localtime.stop()

        # This is a hacky way to clear the cache, since we don't create a filesystem object
        del xds
        if isdir(self.cache_storage(0)):
            rmtree(self.cache_storage(0), ignore_errors=True)



    @staticmethod
    def fv3_path(dates, forecast_hours, file_prefixes):
        """This is passed to :class:`FV3Dataset`, and it generates the paths to read from for the given inputs

        Note:
            With simplecache it's not clear where the cached files go, and they
            do not clear until the process is done running (maybe?) which can file up a filesystem easily.
            So we use filecache instead.

        Args:
            dates (Iterable[datetime]): with the DA cycles to read from
            forecast_hours (List[int]): with the forecast hours to grab ... note here we assume [0, 3] ... but don't enforce it here
            file_prefixes (List[str]): e.g. ["sfg_", "bfg_"]

        Returns:
            list_of_paths (List[str]): see example

        Example:
            >>> mover = ReplayMover( ... )
            >>> mover.cached_path(
                    dates=[datetime(1994,1,1,0), datetime(1994,1,1,6)],
                    forecast_hours=[0, 3],
                    file_prefixes=["sfg_", "bfg_"],
                )
                ["filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010100/sfg_1994010100_fhr00_control,
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010100/sfg_1994010100_fhr03_control
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010100/bfg_1994010100_fhr00_control,
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010100/bfg_1994010100_fhr03_control
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010106/sfg_1994010106_fhr00_control,
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010106/sfg_1994010106_fhr03_control
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010106/bfg_1994010106_fhr00_control,
                "filecache::s3://noaa-ufs-gefsv13replay-pds.s3.amazonaws.com/1deg/1994/01/1994010106/bfg_1994010106_fhr03_control]
        """

        upper = "filecache::s3://noaa-ufs-gefsv13replay-pds/1deg"
        dates = [dates] if not isinstance(dates, Iterable) else dates

        files = []
        for date in dates:
            this_dir = f"{date.year:04d}/{date.month:02d}/{date.year:04d}{date.month:02d}{date.day:02d}{date.hour:02d}"
            for fp in file_prefixes:
                for fhr in forecast_hours:
                    this_file = join(this_dir, f"{fp}{date.year:04d}{date.month:02d}{date.day:02d}{date.hour:02d}_fhr{fhr:02d}_control")
                    files.append(this_file)
        return [join(upper, this_file) for this_file in files]

    @staticmethod
    def mom6_path(cycles, forecast_hours, file_prefixes):
        """This is passed to :class:`MOM6Dataset`, and it generates the paths to read from for the given inputs

        Note:
            With simplecache it's not clear where the cached files go, and they
            do not clear until the process is done running (maybe?) which can file up a filesystem easily.
            So we use filecache instead.

        Args:
            dates (Iterable[datetime]): with the DA cycles to read from
            forecast_hours (List[int]): with the forecast hours to grab ... note here we assume [0, 3] ... but don't enforce it here
            file_prefixes (List[str]): e.g. ["sfg_", "bfg_"]

        Returns:
            list_of_paths (List[str]): see example

        Example:
            >>> mover = ReplayMover( ... )
            >>> mover.cached_path(
                    dates=[datetime(1994,1,1,0), datetime(1994,1,1,6)],
                    forecast_hours=[0],
                    file_prefixes=["ocn_"],
                )
				['filecache::s3://noaa-ufs-gefsv13replay-pds/1deg/1994/01/1994010100/ocn_1994_01_01_00.nc',
                'filecache::s3://noaa-ufs-gefsv13replay-pds/1deg/1994/01/1994010106/ocn_1994_01_01_06.nc']
        """
        upper = "filecache::s3://noaa-ufs-gefsv13replay-pds/1deg"
        cycles = [cycles] if not isinstance(cycles, list) else cycles

        files = []
        for cycle in cycles:
            this_dir = f"{cycle.year:04d}/{cycle.month:02d}/{cycle.year:04d}{cycle.month:02d}{cycle.day:02d}{cycle.hour:02d}"

            for fp in file_prefixes:
                for fhr in forecast_hours:
                    this_date = cycle+timedelta(hours=fhr)
                    this_file = f"{this_dir}/{fp}{this_date.year:04d}_{this_date.month:02d}_{this_date.day:02d}_{this_date.hour:02d}.nc"
                    files.append(this_file)

        return [join(upper, this_file) for this_file in files]


    @staticmethod
    def npdate2datetime(npdate):
        """Convert numpy.datetime64 to datetime.datetime"""
        if not isinstance(npdate, Iterable):
            return datetime(
                    year=int(npdate.dt.year),
                    month=int(npdate.dt.month),
                    day=int(npdate.dt.day),
                    hour=int(npdate.dt.hour),
                )
        else:
            return [datetime(
                    year=int(t.dt.year),
                    month=int(t.dt.month),
                    day=int(t.dt.day),
                    hour=int(t.dt.hour),
                )
                for t in npdate]


    @staticmethod
    def remove_time(xds):
        """Remove time from dataset"""
        single = xds.isel(time=0)
        for key in xds.data_vars:
            if "time" in key:
                del single[key]

        for key in xds.coords:
            if "time" in key:
                del single[key]
        return single

    def add_time_coords(self, xds, time2cftime):
        """add ftime and cftime to container

        This is a bit dirty, passing a static method from another class as a function arg... so it goes.
        """

        iau_offset = -6 if self.component == "fv3" else 0
        repeater = tuple(np.timedelta64(timedelta(hours=fhr-iau_offset)) for fhr in self.forecast_hours)
        ftime = np.array(
            [repeater for _ in range(len(xds["time"])//len(repeater))]
        ).flatten()

        xds["ftime"] = xr.DataArray(
                ftime,
                coords=xds["time"].coords,
                dims=xds["time"].dims,
                attrs={
                    "axis": "T",
                    "description": "time passed since forecast initialization",
                    "long_name": "forecast_time",
                    },
                )
        xds["cftime"] = xr.DataArray(
                time2cftime(xds["time"]),
                coords=xds["time"].coords,
                dims=xds["time"].dims,
                attrs={
                    "calendar_type": "JULIAN",
                    "cartesian_axis": "T",
                    "long_name": "time",
                    },
                )
        xds = xds.set_coords(["ftime", "cftime"])
        return xds


class ReplayMoverQuarterDegree(ReplayMover1Degree):

    def __init__(
        self,
        n_jobs,
        config_filename,
        component="fv3",
        storage_options=None,
        main_cache_path=f"/contrib/Tim.Smith/tmp-replay/0.25-degree",
    ):
        super().__init__(
            n_jobs=n_jobs,
            config_filename=config_filename,
            component=component,
            storage_options=storage_options,
            main_cache_path=main_cache_path,
        )

    @staticmethod
    def fv3_path(dates, forecast_hours, file_prefixes):
        upper = "filecache::s3://noaa-ufs-gefsv13replay-pds"
        dates = [dates] if not isinstance(dates, Iterable) else dates

        files = []
        for date in dates:
            this_dir = f"{date.year:04d}/{date.month:02d}/{date.year:04d}{date.month:02d}{date.day:02d}{date.hour:02d}"
            for fp in file_prefixes:
                for fhr in forecast_hours:
                    this_file = join(this_dir, f"{fp}{date.year:04d}{date.month:02d}{date.day:02d}{date.hour:02d}_fhr{fhr:02d}_control")
                    files.append(this_file)
        return [join(upper, this_file) for this_file in files]

    @staticmethod
    def mom6_path(cycles, forecast_hours, file_prefixes):
        upper = "filecache::s3://noaa-ufs-gefsv13replay-pds"
        cycles = [cycles] if not isinstance(cycles, list) else cycles

        files = []
        for cycle in cycles:
            this_dir = f"{cycle.year:04d}/{cycle.month:02d}/{cycle.year:04d}{cycle.month:02d}{cycle.day:02d}{cycle.hour:02d}"
            for fp in file_prefixes:
                for fhr in forecast_hours:
                    this_date = cycle+timedelta(hours=fhr)
                    this_file = f"{this_dir}/{fp}{this_date.year:04d}_{this_date.month:02d}_{this_date.day:02d}_{this_date.hour:02d}.nc"
                    files.append(this_file)
        return [join(upper, this_file) for this_file in files]



import logging
from typing import Optional
import yaml

import pandas as pd
import xarray as xr

from ufs2arco.sources import Source

logger = logging.getLogger("ufs2arco")


class AWSAORC(Source):
    """
    NOAA Analysis of Record for Calibration (AORC) Dataset available through AWS

    For more information, check out `this page <https://aws.amazon.com/marketplace/pp/prodview-m2sp7gsk5ts6s#resources>`_
    """

    sample_dims = ("time",)
    horizontal_dims = ("latitude", "longitude")
    static_vars = tuple()

    @property
    def available_variables(self) -> tuple:
        return tuple(self._xds.data_vars)

    @property
    def available_levels(self) -> tuple:
        return None

    def __init__(
        self,
        time: dict,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
        slices: Optional[dict] = None,
    ) -> None:

        self.time = pd.date_range(**time)

        # open first year to query variables and rename
        self._xds = self._open_and_rename(self.time[0])

        # parent class checks if variables and levels are legit
        super().__init__(
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
            slices=slices,
        )

        # now subsample the dataset
        self._xds = self._subsample(self._xds)


    def open_sample_dataset(
        self,
        dims: dict,
        open_static_vars: bool,
        cache_dir: Optional[str] = None,
    ) -> xr.Dataset:

        xds = self._open_and_rename(**dims)
        xds = self._subsample(xds)
        xds = xds.sel({k: [v] for k, v in dims.items()})
        osv = open_static_vars or self._open_static_vars(dims)
        selection = list(self.variables) if osv else list(self.dynamic_vars)
        xds = xds[selection].copy().load()
        return xds


    def _open_and_rename(self, time: pd.Timestamp) -> xr.Dataset:

        xds = xr.open_zarr(
            self._build_uri(time),
            storage_options={"s3": {"anon": True}},
            decode_timedelta=True,
        )
        return xds.rename(self.rename)


    def _subsample(self, xds: xr.Dataset) -> xr.Dataset:

        xds = xds[self.variables]
        if self.levels is not None:
            xds = xds.sel(level=self.levels, **self._level_sel_kwargs)
        xds = self.apply_slices(xds)
        return xds


    def _build_uri(self, time: pd.Timestamp) -> str:
        return f"s3://noaa-nws-aorc-v1-1-1km/{time.year}.zarr"




import logging
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr

from ufs2arco.sources import Source, NOAAGribForecastData

logger = logging.getLogger("ufs2arco")

class AWSGEFSArchive(NOAAGribForecastData, Source):
    """Access NOAA's forecast archive from the Global Ensemble Forecast System (GEFS) at s3://noaa-gefs-pds."""

    sample_dims = ("t0", "fhr", "member")
    horizontal_dims = ("latitude", "longitude")
    file_suffixes = ("a", "b")
    static_vars = ("lsm", "orog")

    @property
    def available_levels(self) -> tuple:
        return (
            10, 20, 30, 50, 70,
            100, 150, 200, 250, 300, 350, 400, 450,
            500, 550, 600, 650, 700, 750, 800, 850,
            900, 925, 950, 975, 1000,
        )

    @property
    def rename(self) -> dict:
        return {
            "time": "t0",
            "step": "lead_time",
            "isobaricInhPa": "level",
        }

    def __init__(
        self,
        t0: dict,
        fhr: dict,
        member: dict,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
        slices: Optional[dict] = None,
        accum_hrs: Optional[dict] = None,
    ) -> None:
        """
        Args:
            t0 (dict): Dictionary with start and end times for initial conditions, and e.g. "freq=6h". All options get passed to ``pandas.date_range``.
            fhr (dict): Dictionary with 'start', 'end', and 'step' forecast hours.
            member (dict): Dictionary with 'start', 'end', and 'step' ensemble members.
            variables (list, tuple, optional): variables to grab
            levels (list, tuple, optional): vertical levels to grab
            use_nearest_levels (bool, optional): if True, all level selection with
                ``xarray.Dataset.sel(level=levels, method="nearest")``
            slices (dict, optional): either "sel" or "isel", with slice, passed to xarray
            accum_hrs (dict, optional): period in hours over which to accumulate accumulated variables, e.g.
                {"accum_tp": 1}
                would be the same as passing filter_by_keys={"stepRange": "5-6"}
                when reading fhr 6 data using xarray and cfgrib
        """
        self.t0 = pd.date_range(**t0)
        self.fhr = np.arange(fhr["start"], fhr["end"] + 1, fhr["step"])
        self.member = np.arange(member["start"], member["end"] + 1, member["step"])
        super().__init__(
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
            slices=slices,
            accum_hrs=accum_hrs,
        )


    def _build_path(
        self,
        t0: pd.Timestamp,
        member: int,
        fhr: int,
        file_suffix: str,
    ) -> str:
        """
        Build the file path to a GEFS GRIB file on AWS.

        Args:
            t0 (pd.Timestamp): The initial condition timestamp.
            member (int): The ensemble member ID.
            fhr (int): The forecast hour.
            file_suffix (str): For GEFS, either 'a' or 'b'.

        Returns:
            str: The constructed file path.
        """
        c_or_p = "c" if member == 0 else "p"
        bucket = f"s3://noaa-gefs-pds"
        outer = f"gefs.{t0.year:04d}{t0.month:02d}{t0.day:02d}/{t0.hour:02d}"
        # Thanks to Herbie for figuring these out
        if t0 < pd.Timestamp("2018-07-27"):
            middle = ""
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{file_suffix}f{fhr:03d}"
        elif t0 < pd.Timestamp("2020-09-23T12"):
            middle = f"pgrb2{file_suffix}/"
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{file_suffix}f{fhr:02d}"
        else:
            middle = f"atmos/pgrb2{file_suffix}p5/"
            fname = f"ge{c_or_p}{member:02d}.t{t0.hour:02d}z.pgrb2{file_suffix}.0p50.f{fhr:03d}"
        fullpath = f"filecache::{bucket}/{outer}/{middle}{fname}"
        logger.debug(f"{self.name}._build_path: reading {fullpath}")
        return fullpath



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import datetime
sys.path.insert(0, os.path.abspath("../"))

project = 'ufs2arco'
copyright = f"2023-{datetime.datetime.now().year}, ufs2arco developers"
author = 'ufs2arco developers'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "nbsphinx",
]

numpydoc_show_class_members = False
napolean_google_docstring = True
napolean_numpy_docstring = False

templates_path = ['_templates']
exclude_patterns = ["archive/*.ipynb"]

napoleon_custom_sections = [
    ("Returns", "params_style"),
    ("Sets Attributes", "params_style"),
    ("Assumptions", "notes_style"),
    ("Required Fields in Config", "params_style"),
    ("Optional Fields in Config", "params_style"),
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']

html_theme_options = {
    "repository_url": "https://github.com/NOAA-PSL/ufs2arco",
    "use_repository_button": True,
}



import itertools
import re
from collections import defaultdict

import numpy as np
import xarray as xr

def batched(iterable, n):
    """A homegrown version of itertools.batched, taken
    directly from `itertools docs <https://docs.python.org/3/library/itertools.html#itertools.batched>`_
    for earlier versions of itertools

    Args:
        iterable (Iterable): loop over this
        n (int): batch size

    Returns:
        batch (Iterable): a batch of size n from the iterable
    """
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


def expand_anemoi_dataset(ads: xr.Dataset, name: str, variable_names: list[str]) -> xr.Dataset:
    """Expand the anemoi dataset "variable" dimension to individual variables.

    Converts a DataArray within the anemoi dataset,
    with any subset of dimensions [time, variable, ensemble, cell] into an xarray.Dataset
    where each variable has dimensions [time, cell] for 2D variables, or [time, level, cell] for 3D variables.

    Note that this renames "latitudes" to "latitude", "longitudes" to "longitude", and swaps the logical "time"
    dimension with "dates", and renames "dates" to "time".

    Example:
        >>> import xarray as xr
        >>> from ufs2arco.utils import expand_anemoi_dataset
        >>> ads = xr.open_zarr("/path/to/my-anemoi-dataset.zarr")
        >>> the_data = expand_anemoi_dataset(ads, "data", ads.attrs["variables"])
        >>> maximum_values = expand_anemoi_dataset(ads, "maximum", ads.attrs["variables"])

    Parameters:
        ads (xr.Dataset): anemoi dataset, opened with xarray
        name (str): name of the array within the anemoi dataset to expand, e.g. "data" or any of the statistics ("maximum", "minimum", "squares", "mean", etc)
        variable_names (list[str]): A list of variable names corresponding to the 'variable' dimension.

    Returns:
        xr.Dataset: A dataset with each variable as a [time, cell] DataArray.
    """
    xda = ads[name]
    if xda.sizes['variable'] != len(variable_names):
        raise ValueError("Length of variable_names does not match the size of the 'variable' dimension.")

    # Use regex to detect base name and level
    level_pattern = re.compile(r"^(?P<base>[a-zA-Z0-9]+)_(?P<level>\d+)$")
    level_groups = defaultdict(list)
    flat_vars = []

    for name in variable_names:
        parts = name.rsplit("_", 1)
        if len(parts) > 1 and parts[-1].isdigit():
            base = parts[0]
            level = int(parts[-1])
            level_groups[base].append((level, name))
        else:
            flat_vars.append(name)

    dsdict = {}

    # Add grouped variables with levels
    for base, level_name_pairs in level_groups.items():

        dalist = []
        for level, level_name in level_name_pairs:
            this_2d_array = xda.sel(variable=variable_names.index(level_name))
            this_2d_array = this_2d_array.drop_vars("variable")
            this_2d_array = this_2d_array.expand_dims({"level": [level]})
            dalist.append(this_2d_array)

        dsdict[base] = xr.concat(dalist, dim="level")

    # Add flat variables
    for name in flat_vars:
        dsdict[name] = xda.sel(variable=variable_names.index(name)).drop_vars("variable")

    xds = xr.Dataset({k: dsdict[k] for k in sorted(dsdict)})
    for key in xds.data_vars:
        dims = tuple(d for d in ("time", "ensemble", "level", "cell") if d in xds[key].dims)
        xds[key] = xds[key].transpose(*dims)

    # rename stuff, set coordinates, swap dims
    for key in ["dates", "latitudes", "longitudes"]:
        xds[key] = ads[key]
        xds = xds.set_coords(key)
    xds = xds.rename({"latitudes": "latitude", "longitudes": "longitude"})
    xds = xds.swap_dims({"time": "dates"}).drop_vars("time").rename({"dates": "time"})
    return xds


def convert_anemoi_inference_dataset(xds: xr.Dataset):
    """Convert the output from anemoi-inference to a multivariate xarray dataset.

    Stack each variable separated by vertical level to a single variable with a level dimension

    Note that this also renames dimension names "values" -> "cell" so that it is consistent with
    the anemoi dataset naming convention, and to avoid issues with xarray.DataArray.values.

    Example:
        >>> import xarray as xr
        >>> from ufs2arco.utils import convert_anemoi_inference_dataset
        >>> ids = xr.open_dataset("/path/to/anemoi_inference.nc", decode_timedelta=True, chunks="auto")
        >>> convert_anemoi_inference_dataset(ids)

    Args:
        xds (xr.Dataset): result from anemoi inference run, stored as netcdf

    Returns:
        result (xr.Dataset): with vertical levels stacked and "values" renamed to "cell"
    """

    # Use regex to detect base name and level
    level_pattern = re.compile(r"^(?P<base>[a-zA-Z0-9]+)_(?P<level>\d+)$")
    level_groups = defaultdict(list)
    flat_vars = []

    for name in xds.data_vars:
        parts = name.rsplit("_", 1)
        if len(parts) > 1 and parts[-1].isdigit():
            base = parts[0]
            level = int(parts[-1])
            level_groups[base].append((level, name))
        else:
            flat_vars.append(name)

    dsdict = {}

    # Add grouped variables with levels
    for base, level_name_pairs in level_groups.items():

        dalist = []
        for level, level_name in level_name_pairs:
            this_2d_array = xds[level_name].expand_dims({"level": [level]})
            attrs = xds[level_name].attrs.copy()
            dalist.append(this_2d_array)

        dsdict[base] = xr.concat(dalist, dim="level")
        dsdict[base].attrs = attrs

    # Add flat variables
    for name in flat_vars:
        dsdict[name] = xds[name]
        dsdict[name].attrs = xds[name].attrs.copy()

    result = xr.Dataset({k: dsdict[k] for k in sorted(dsdict)}, attrs=xds.attrs)
    result = result.set_coords(["latitude", "longitude"])
    result = result.rename({"values": "cell"})
    for key in result.data_vars:
        dims = tuple(d for d in ("time", "ensemble", "level", "cell") if d in result[key].dims)
        result[key] = result[key].transpose(*dims)
    return result



try:
    from mpi4py import MPI
except:
    pass

import os
import logging
import yaml
from datetime import datetime

import numpy as np
import xarray as xr
import pandas as pd
import zarr

from ufs2arco.log import setup_simple_log
from ufs2arco.mpi import MPITopology, SerialTopology
import ufs2arco.sources
from ufs2arco.transforms import Transformer
import ufs2arco.targets
from ufs2arco.datamover import DataMover, MPIDataMover

logger = logging.getLogger("ufs2arco")

class Driver:
    """A class to manage data movement.

    Attributes:
        config (dict): Configuration dictionary loaded from the YAML file.
        Source, Transformer, Target
        Mover (Type[DataMover] | Type[MPIDataMover]): The data mover class (DataMover or MPIDataMover).

    Methods:
        __init__(config_filename: str): Initializes the Driver object with configuration from the specified YAML file.
        use_mpi: Returns whether MPI should be used based on the configuration.
        source_kwargs: Returns the arguments for initializing the source dataset.
        target_kwargs: Returns the arguments for initializing the target dataset.
        mover_kwargs: Returns the arguments for initializing the mover.
        run(overwrite: bool = False): Runs the data movement process, managing the source, target transformations, and mover.
    """

    required_sections = (
        "mover",
        "directories",
        "source",
        "target",
    )

    recognized_sections = (
        "mover",
        "directories",
        "source",
        "transforms",
        "target",
        "attrs",
    )

    def __init__(self, config_filename: str):
        """Initializes the Driver object with configuration from the specified YAML file.

        Note:
            Initialization does very little, since we need to start the run in order to setup the MPI or Serial Topology that orchestrates everything. This is used for logging. After that, we can initialize the source, target, etc.

        Args:
            config_filename (str): Path to the YAML configuration file.

        Raises:
            AssertionError: If required sections or keys are missing in the configuration.
            NotImplementedError: If a source, target, or mover is not recognized.
        """

        with open(config_filename, "r") as f:
            self.config = yaml.safe_load(f)

        for key in self.required_sections:
            assert key in self.config, \
                f"Driver.__init__: could not find '{key}' section in yaml"

        unrecognized = []
        for key in self.config.keys():
            if key not in self.recognized_sections:
                unrecognized.append(key)
        if len(unrecognized) > 0:
            raise KeyError(
                f"Driver.__init__: Unrecognized config sections: {unrecognized}. The following are recognized: {self.recognized_sections}")

        # directories
        dirs = self.config["directories"]
        for key in ["zarr", "cache"]:
            assert key in dirs, \
                f"Driver.__init__: could not find '{key}' in 'directories' section in yaml"

        for key in ["cache_dir"]:
            assert key not in self.config["mover"], \
                f"Driver.__init__: '{key}' not allowed in 'mover' section of yaml"

        # make sure chunks are here
        for key in ["chunks"]:
            assert key in self.config["target"], \
                f"Driver.__init__: could not find '{key}' in 'target' section of yaml"


    def _init_topo(self, runtype: str):
        """Initialize MPITopology or SerialTopology, which sets up the logging"""

        mover_kwargs = self.mover_kwargs.copy()

        log_dir = os.path.expandvars(self.config["directories"]["logs"])

        if runtype != "create":
            if log_dir.endswith("/"):
                log_dir = log_dir[:-1]
            log_dir += f"-{runtype}"

        if self.use_mpi:
            self.topo = MPITopology(log_dir=log_dir)

        else:
            self.topo = SerialTopology(log_dir=log_dir)


    def _init_source(self):
        """Check and Initialize the Source Dataset and Transformer"""

        name = self.config["source"]["name"].lower()
        if name not in ufs2arco.sources._recognized:
            raise NotImplementedError(f"Driver._init_source_and_transforms: unrecognized data source {name}. Must be one of {ufs2arco.sources._recognized}")
        SourceDataset = getattr(ufs2arco.sources, ufs2arco.sources._recognized[name])

        self.source = SourceDataset(**self.source_kwargs)

    def _init_transformer(self):
        """create the transformer"""

        if "transforms" in self.config.keys():
            self.transformer = Transformer(options=self.config["transforms"])
        else:
            self.transformer = None


    def _init_target(self):
        """Check and Initialize the Target Dataset format"""

        name = self.config["target"].get("name", "base").lower()
        if name in ("forecast", "analysis", "base"):
            TargetDataset = ufs2arco.targets.Target
        elif name == "anemoi":
            TargetDataset = ufs2arco.targets.Anemoi
        else:
            raise NotImplementedError(f"Driver._init_targets: only 'base' and 'anemoi' are implemented")

        self.target = TargetDataset(source=self.source, **self.target_kwargs)


    def _init_mover(self):
        """Initialize the DataMover"""

        kwargs = self.mover_kwargs.copy()
        if self.use_mpi:
            Mover = MPIDataMover
            kwargs["mpi_topo"] = self.topo
        else:
            Mover = DataMover

        self.mover = Mover(source=self.source, target=self.target, transformer=self.transformer, **kwargs)


    @property
    def use_mpi(self) -> bool:
        """Determines if MPI is to be used based on the mover configuration.

        Returns:
            bool: True if MPI is to be used, False otherwise.
        """
        return self.config["mover"]["name"].lower() == "mpidatamover"

    @property
    def source_kwargs(self) -> dict:
        """Returns the arguments for initializing the source dataset.

        Returns:
            dict: The source dataset initialization arguments.
        """
        kw = {key: val for key, val in self.config["source"].items() if key != "name"}
        return kw

    @property
    def target_kwargs(self) -> dict:
        """Returns the arguments for initializing the target dataset.

        Returns:
            dict: The target dataset initialization arguments.
        """
        kw = {key: val for key, val in self.config["target"].items() if key != "name"}
        kw["store_path"] = os.path.expandvars(self.config["directories"]["zarr"])
        return kw

    @property
    def mover_kwargs(self) -> dict:
        """Returns the arguments for initializing the data mover.

        Returns:
            dict: The mover initialization arguments.
        """
        kw = {key: val for key, val in self.config["mover"].items() if key != "name"}

        # optional
        if "start" not in kw.keys():
            kw["start"] = 0

        # enforced options
        kw["cache_dir"] = os.path.expandvars(self.config["directories"]["cache"])
        return kw


    @property
    def store_path(self) -> str:
        return self.target.store_path


    def setup(self, runtype: str):
        self._init_topo(runtype)
        self._init_source()
        self._init_transformer()
        self._init_target()
        self._init_mover()


    def write_container(self, overwrite):
        """Write empty zarr store, to be filled with data"""

        if self.topo.is_root:
            cds = self.mover.create_container()

            kwargs = {"mode": "w"} if overwrite else {}
            logger.info(f"Driver.write_container: storing container at {self.store_path}\n{cds}\n")
            cds.to_zarr(self.store_path, compute=False, **kwargs)
            logger.info("Driver.write_container: Done storing container.\n")

        self.topo.barrier()


    def run(self, overwrite: bool = False):
        """Runs the data movement process, managing the datasets and mover.

        This method sets up the datasets, creates the container, and loops through
        batches to move data to the specified store path (Zarr format).

        Args:
            overwrite (bool, optional): Whether to overwrite the existing container.
                Defaults to False.
        """
        self.setup(runtype="create")

        # create container, only if mover start is not 0
        if self.mover.start == 0:
            self.write_container(overwrite=overwrite)

        # loop through batches
        n_batches = len(self.mover)
        missing_dims = []
        for batch_idx in range(self.mover.start, n_batches):

            xds = next(self.mover)


            # xds is None if MPI rank looks for non existent indices (i.e., last batch scenario)
            # len(xds) == 0 if we couldn't find the file we were looking for
            has_content = xds is not None and len(xds) > 0
            if has_content:

                xds = xds.reset_coords(drop=True)
                region = self.mover.find_my_region(xds)
                xds.to_zarr(self.store_path, region=region)
                self.mover.clear_cache(batch_idx)

            elif xds is not None:

                # we couldn't find the file, keep track of it
                batch_indices = self.mover.get_batch_indices(batch_idx)
                for these_dims in batch_indices:
                    missing_dims.append(these_dims)

            logger.info(f"Done with batch {batch_idx+1} / {n_batches}")

        self.topo.barrier()
        logger.info(f"Done moving the data\n")

        self.report_missing_data(missing_dims)
        self.target.finalize(self.topo)
        self.finalize_attributes()

    def patch(self):

        self.setup(runtype="patch")
        missing_dims = _open_patch_yaml(self.get_missing_data_path(self.store_path))

        logger.info(f"Starting patch workflow with missing_dims\n{missing_dims}\n")

        # hacky for now, reset the mover's sample_indices to be the missing dims
        self.mover.sample_indices = missing_dims
        self.mover.restart(idx=0)

        missing_again = list()
        n_batches = len(self.mover)
        for batch_idx, xds in enumerate(self.mover):

            # xds is None if MPI rank looks for non existent indices (i.e., last batch scenario)
            # len(xds) == 0 if we couldn't find the file we were looking for
            has_content = xds is not None and len(xds) > 0
            if has_content:

                xds = xds.reset_coords(drop=True)
                region = self.mover.find_my_region(xds)
                xds.to_zarr(self.store_path, region=region)
                self.mover.clear_cache(batch_idx)

            elif xds is not None:

                batch_indices = self.mover.get_batch_indices(batch_idx)
                for these_dims in batch_indices:
                    missing_again.append(these_dims)

            logger.info(f"Done with batch {batch_idx+1} / {n_batches}")

        self.topo.barrier()
        logger.info(f"Done moving the data\n")

        self.report_missing_data(missing_again)
        self.target.finalize(self.topo)
        self.finalize_attributes()


    def finalize_attributes(self):

        logger.info(f"Storing the recipe and anything from the 'attrs' section in zarr store attributes")
        if self.topo.is_root:
            zds = zarr.open(self.store_path, mode="a")
            zds.attrs["recipe"] = self.config
            if "attrs" in self.config.keys():
                for key, val in self.config["attrs"].items():
                    zds.attrs[key] = val

            zds.attrs["latest_write_timestamp"] = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            # just in case
            zarr.consolidate_metadata(self.store_path)
        self.topo.barrier()
        logger.info(f"🚀🚀🚀 Dataset is ready for launch at: {self.store_path}")


    def get_missing_data_path(self, store_path) -> str:
        directory, zstore = os.path.split(store_path)
        return f"{directory}/missing.{zstore}.yaml"

    def report_missing_data(self, missing_dims):

        missing_dims = self.topo.gather(missing_dims)

        if self.topo.is_root:
            # collect the list of lists into one list of dicts
            missing_dims = [item for sublist in missing_dims for item in sublist]

            if len(missing_dims) > 0:
                # sort it by t0 or time
                # then convert numpy or pandas types to int/str etc
                missing_dims = sorted(missing_dims, key=_get_time)
                missing_dims = [_convert_types_to_yaml(d) for d in missing_dims]

                self.target.handle_missing_data(missing_dims)

                missing_data_yaml = self.get_missing_data_path(self.store_path)
                with open(missing_data_yaml, "w") as f:
                    yaml.dump(missing_dims, stream=f)

                logger.warning(f"⚠️  Some data are missing.")
                logger.warning(f"⚠️  The missing dimension combos, i.e., {self.source.sample_dims}")
                logger.warning(f"⚠️  were written to: {missing_data_yaml}")
                logger.warning(f"If you know the files are actually available, You can run")
                logger.warning(f"\tpython -c 'import ufs2arco; ufs2arco.Driver(\"/path/to/your/original/recipe.yaml\").patch()'")
                logger.warning(f"to try getting the data again\n")

# some utilities for handling missing data
def _get_time(d):
    return d.get("t0", d.get("time", None))

def _convert_types_to_yaml(d):
    d = d.copy()
    # Convert pd.Timestamp to string
    if "t0" in d and isinstance(d["t0"], pd.Timestamp):
        d["t0"] = str(d["t0"])
    if "time" in d and isinstance(d["time"], pd.Timestamp):
        d["time"] = str(d["time"])
    # Convert numpy integers to Python int
    for key in ["fhr", "member"]:
        if key in d and isinstance(d[key], np.integer):
            d[key] = int(d[key])
    return d

def _open_patch_yaml(yamlpath):

    with open(yamlpath, "r") as f:
        missing_dims = yaml.safe_load(f)
    # Convert string -> pd.Timestamp
    for i, d in enumerate(missing_dims):
        if "t0" in d and isinstance(d["t0"], str):
            missing_dims[i]["t0"] = pd.Timestamp(d["t0"])
        if "time" in d and isinstance(d["time"], str):
            missing_dims[i]["time"] = pd.Timestamp(d["time"])
    return missing_dims



import argparse
import yaml
from ufs2arco.driver import Driver
from ufs2arco.multidriver import MultiDriver

def main():
    parser = argparse.ArgumentParser(
        description="Run the ufs2arco workflow with a given YAML recipe.",
    )
    parser.add_argument(
        "yaml_file",
        type=str,
        help="Path to the YAML recipe file.",
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Pass this flag to overwrite an existing zarr store in the specified location",
    )
    args = parser.parse_args()

    with open(args.yaml_file, "r") as f:
        config = yaml.safe_load(f)

    if "multisource" in config.keys():
        MultiDriver(args.yaml_file).run(overwrite=args.overwrite)
    else:
        Driver(args.yaml_file).run(overwrite=args.overwrite)

if __name__ == "__main__":
    main()



try:
    from mpi4py import MPI
except:
    pass

import os
import logging
import yaml
from datetime import datetime

import numpy as np
import xarray as xr
import pandas as pd
import zarr

from ufs2arco.log import setup_simple_log
from ufs2arco.mpi import MPITopology, SerialTopology
import ufs2arco.sources
from ufs2arco.transforms import Transformer
from ufs2arco import targets
from ufs2arco.datamover import DataMover, MPIDataMover

from ufs2arco.driver import Driver

logger = logging.getLogger("ufs2arco")

class MultiDriver(Driver):
    """A class to manage data movement, with multiple sources.

    Note:
        For now, all "sources" must come from the same dataset, e.g. we can't mix and match GFS and ERA5.
        There are several reasons for this:
            * Right now, this is baked into how the caching is done. We only clear one data mover's cache, and all movers use the same cache directory.
            * It's not clear how different source types (e.g. reanalysis and forecast, or deterministic and ensemble) would combine, given the different coordinates. This would make combining the datasets, and finding the specific spot in the zarr dataset (Mover.find_my_region) much more complicated.


    Attributes:
        config (dict): Configuration dictionary loaded from the YAML file.
        Source, Transformer, Target
        Mover (Type[DataMover] | Type[MPIDataMover]): The data mover class (DataMover or MPIDataMover).

    Methods:
        __init__(config_filename: str): Initializes the Driver object with configuration from the specified YAML file.
        use_mpi: Returns whether MPI should be used based on the configuration.
        source_kwargs: Returns the arguments for initializing the source dataset.
        target_kwargs: Returns the arguments for initializing the target dataset.
        mover_kwargs: Returns the arguments for initializing the mover.
        run(overwrite: bool = False): Runs the data movement process, managing the source, target transformations, and mover.
    """

    required_sections = (
        "mover",
        "directories",
        "multisource",
        "target",
    )

    recognized_sections = (
        "mover",
        "directories",
        "multisource",
        "transforms",
        "target",
        "attrs",
    )

    @property
    def source_kwargs(self) -> dict:
        """Returns the arguments for initializing the source dataset.

        Returns:
            dict: The source dataset initialization arguments.
        """
        return [
            {key: val for key, val in local_config["source"].items() if key != "name"}
            for local_config in self.config["multisource"]
        ]

    @property
    def source(self):
        return self.sources[0]

    @property
    def target(self):
        return self.targets[0]

    @property
    def mover(self):
        return self.movers[0]


    def _init_source(self):
        """Check and Initialize each Source Dataset"""

        # First, make sure the names are legit
        SourceDatasets = list()
        for local_config in self.config["multisource"]:

            name = local_config["source"]["name"].lower()
            if name not in ufs2arco.sources._recognized:
                raise NotImplementedError(f"Driver.__init__: unrecognized data source {name}. Must be one of {ufs2arco.sources._recognized}")

            SourceDatasets.append( getattr(ufs2arco.sources, ufs2arco.sources._recognized[name]) )

        try:
            assert all(local_config["source"]["name"].lower() == name for local_config in self.config["multisource"])
        except:
            raise NotImplementedError("For multisource workflows, all sources have to come from the same dataset. Can't mix and match")

        # Now initialize the actual objects
        self.sources = list()
        for LocalSource, local_kwargs in zip(SourceDatasets, self.source_kwargs):
            self.sources.append( LocalSource(**local_kwargs) )


    def _init_transformer(self):
        """Create the transformer

        For the multi source, we need a transformer for each object.
        """

        self.transformers = list()
        common_transforms = self.config.get("transforms", {})
        for local_config in self.config["multisource"]:
            unique_transforms = local_config.get("transforms", {})
            all_transforms = {**unique_transforms.copy(), **common_transforms.copy()}
            if len(all_transforms) > 0:
                transformer = Transformer(options=all_transforms)
            else:
                transformer = None
            self.transformers.append(transformer)


    def _init_target(self):
        """Make a copy of the same target format for each source

        Note that we only compute any forcings with the first data source
        """

        name = self.config["target"].get("name", "base").lower()
        try:
            assert name == "anemoi"
        except:
            raise NotImplementedError("Driver._init_target: multisource workflows currently only work with Anemoi Targets.")
        self.targets = list()
        kwargs = self.target_kwargs.copy()
        for source in self.sources:
            self.targets.append(
                ufs2arco.targets.Anemoi(
                    source=source,
                    **kwargs,
                )
            )
            # After the first source, drop "forcings" from target kwargs... no need to compute them more than once
            kwargs.pop("forcings", None)


    def _init_mover(self):

        kwargs = self.mover_kwargs.copy()
        if self.use_mpi:
            Mover = MPIDataMover
            kwargs["mpi_topo"] = self.topo
        else:
            Mover = DataMover

        self.movers = [
            Mover(source=source, target=target, transformer=transformer, **kwargs)
            for source, target, transformer in zip(self.sources, self.targets, self.transformers)
        ]


    def write_container(self, overwrite):
        """Write empty zarr store, to be filled with data"""

        if self.topo.is_root:
            dslist = [mover.create_container() for mover in self.movers]
            cds = self.target.merge_multisource(dslist)

            kwargs = {"mode": "w"} if overwrite else {}
            logger.info(f"Driver.write_container: storing container at {self.store_path}\n{cds}\n")
            cds.to_zarr(self.store_path, compute=False, **kwargs)
            logger.info("Driver.write_container: Done storing container.\n")

        self.topo.barrier()


    def run(self, overwrite: bool = False):
        """Runs the data movement process, managing the datasets and mover.

        This method sets up the datasets, creates the container, and loops through
        batches to move data to the specified store path (Zarr format).

        Args:
            overwrite (bool, optional): Whether to overwrite the existing container.
                Defaults to False.
        """
        self.setup(runtype="create")

        # create container, only if mover start is not 0
        if self.mover.start == 0:
            self.write_container(overwrite=overwrite)

        # loop through batches
        n_batches = len(self.mover)
        missing_dims = []
        for batch_idx in range(self.mover.start, n_batches):

            dslist = list()
            foundit = list()
            for mover in self.movers:

                xds = next(mover)

                # xds is None if MPI rank looks for non existent indices (i.e., last batch scenario)
                # len(xds) == 0 if we couldn't find the file we were looking for
                has_content = xds is not None and len(xds) > 0
                if has_content:
                    foundit.append(True)
                    dslist.append(xds.reset_coords(drop=True))

                elif xds is not None:

                    foundit.append(False)
                    batch_indices = mover.get_batch_indices(batch_idx)
                    for these_dims in batch_indices:
                        missing_dims.append(these_dims)

            # we need both conditionals, since all([]) == True
            if all(foundit) and len(foundit) == len(self.movers):

                mds = self.target.merge_multisource(dslist)

                region = self.mover.find_my_region(mds)
                mds.to_zarr(self.target.store_path, region=region)

            self.mover.clear_cache(batch_idx)

            logger.info(f"Done with batch {batch_idx+1} / {n_batches}")

        self.topo.barrier()
        logger.info(f"Done moving the data\n")

        self.report_missing_data(missing_dims)
        self.target.finalize(self.topo)
        self.finalize_attributes()

    def patch(self):
        raise NotImplementedError



import os
import yaml
from typing import Optional
import numpy as np
import xarray as xr

class Layers2Pressure():
    """A class to interpolate from Lagrangian layers to pressure levels, and also compute layer thickness, etc

    Args:
        ak, bk (np.ndarray, optional): coefficients that define the vertical grid. If not provided, will
            default to the 127 vertical levels defined for NOAA GFS
        pfull, phalf (np.ndarray, optional): if ak and bk are provided, these are not necessary since they
            can be computed from those coefficients. However, these can be provided in case there are slight
            numerical differences between how these values are computed here, versus the grid information
            from an existing dataset.
        level_name, interface_name (str, optional): names to use for the vertical coordinate at
            cell center and interfaces, defaulting to "pfull" and "phalf" as in FV3. However,
            new names can be provided in order to work with datasets that use a different name.
    """
    g = 9.80665     # m / s^2
    Rd = 287.05     # J / kg / K
    Rv = 461.5      # J / kg / K
    q_min = 1e-10   # kg / kg
    z_vir = Rv / Rd - 1.
    xds = {"ak": None, "bk": None, "pfull": None, "phalf": None}

    def __init__(
        self,
        ak: Optional[np.ndarray]=None,
        bk: Optional[np.ndarray]=None,
        pfull: Optional[np.ndarray]=None,
        phalf: Optional[np.ndarray]=None,
        level_name: Optional[str]="pfull",
        interface_name: Optional[str]="phalf",
    ):

        self.level_name = level_name
        self.interface_name = interface_name

        if ak is None and bk is None:
            self.xds = self._get_default_vertical_grid()
        elif (ak is not None and bk is None) or (ak is None and bk is not None):
            raise TypeError("Must provide ak and bk, or neither, not just one.")
        else:

            newxds = self._get_xds(ak, bk, pfull, phalf)
            self.xds = newxds


    def _get_default_vertical_grid(self):

        # read vertical coordinate information
        vcoord_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "replay_vertical_levels.yaml")
        with open(vcoord_path, "r") as f:
            vcoords = yaml.safe_load(f)

        ak = np.array(vcoords["ak"])
        bk = np.array(vcoords["bk"])
        pfull = np.array(vcoords["pfull"])
        return self._get_xds(ak, bk, pfull)


    def _get_xds(
        self,
        ak: np.ndarray,
        bk: np.ndarray,
        pfull: Optional[np.ndarray]=None,
        phalf: Optional[np.ndarray]=None,
    ) -> xr.Dataset:

        ak = ak.values if isinstance(ak, xr.DataArray) else ak
        bk = bk.values if isinstance(bk, xr.DataArray) else bk

        # Compute phalf and pfull if necessary
        if phalf is None:
            phalf = (ak + 100_000 * bk) / 100 # hPa

        if pfull is None:
            pfull = 0.5*phalf[1:] + 0.5*phalf[:-1]


        # Now convert
        phalf = xr.DataArray(
            phalf,
            coords={self.interface_name: phalf},
            dims=(self.interface_name,),
        )
        pfull = xr.DataArray(
            pfull,
            coords={self.level_name: pfull},
            dims=(self.level_name,),
        )

        ak = xr.DataArray(
            ak,
            coords={self.interface_name: phalf},
            dims=(self.interface_name,),
        )
        bk = xr.DataArray(
            bk,
            coords=ak.coords,
            dims=ak.dims,
        )

        xds = xr.Dataset({
            "ak": ak,
            "bk": bk,
            self.level_name: pfull,
            self.interface_name: phalf,
        })
        xds = xds.set_coords(["ak", "bk", self.interface_name, self.level_name])
        return xds

    @property
    def ak(self):
        return self.xds["ak"]

    @property
    def bk(self):
        return self.xds["bk"]

    @property
    def pfull(self):
        return self.xds[self.level_name]

    @property
    def phalf(self):
        return self.xds[self.interface_name]


    def calc_pressure_interfaces(self, pressfc: xr.DataArray) -> xr.DataArray:
        """Compute pressure at vertical grid interfaces

        Args:
            pressfc (xr.DataArray): surface pressure (Pa)

        Returns:
            prsi (xr.DataArray): 3D pressure field (or 4D if input has time) at vertical grid interfaces (Pa)
        """

        return self.ak.astype(pressfc.dtype) + pressfc*self.bk.astype(pressfc.dtype)

    def calc_pressure_thickness(self, prsi: xr.DataArray) -> xr.DataArray:
        """Compute pressure thickness at each vertical level

        Args:
            prsi (xr.DataArray): Pressure at vertical grid interfaces (Pa)

        Returns:
            dpres (xr.DataArray): Pressure thickness at each vertical level (Pa)
        """
        dpres = prsi.diff(self.interface_name, label="lower")
        return self._dphalf_to_pfull(dpres, name="dpres")

    def calc_dlogp(self, prsi: xr.DataArray) -> xr.DataArray:
        """Compute difference of log of interface pressure

        Args:
            prsi (xr.DataArray): Pressure at vertical grid interfaces

        Returns:
            dlogp (xr.DataArray): np.log(prsi).diff("phalf")
        """
        dlogp = np.log(prsi).diff(self.interface_name, label="lower")
        return self._dphalf_to_pfull(dlogp, name="dlogp")

    def calc_delz(
        self,
        pressfc: xr.DataArray,
        temp: xr.DataArray,
        spfh: xr.DataArray,
    ) -> xr.DataArray:
        """Compute a hydrostatic approximation of the layer thickness at each vertical level.

        Args:
            pressfc, temp, spfh (xr.DataArray): surface pressure, temperature, and specific humidity

        Returns:
            delz (xr.DataArray): layer thickness (m)
        """
        prsi = self.calc_pressure_interfaces(pressfc)
        dlogp = self.calc_dlogp(prsi)
        dlogp = dlogp.sel({self.level_name: temp[self.level_name]})
        spfh_thresh = spfh.where(spfh > self.q_min, self.q_min)
        return -self.Rd /  self.g * temp * (1. + self.z_vir*spfh_thresh) * dlogp

    def calc_layer_mean_pressure(
        self,
        pressfc: xr.DataArray,
        temp: xr.DataArray,
        spfh: xr.DataArray,
        delz: xr.DataArray,
    ) -> xr.DataArray:
        """Compute pressure at vertical grid cell center (i.e., layer mean)

        Note:
            If ``delz`` is computed by :meth:`calc_delz`, then a hydrostatic approximation is used. This may be inconsistent with the original simulation/dataset.

        Args:
            pressfc, temp, spfh, delz (xr.DataArray): surface pressure, temperature, specific humidity, and layer thickness

        Returns:
            prsl (xr.DataArray): layer mean thickness
        """

        prsi = self.calc_pressure_interfaces(pressfc)
        dpres = self.calc_pressure_thickness(prsi)
        dpres = dpres.sel({self.level_name: temp[self.level_name]})

        # rTv computed here:
        # https://github.com/NOAA-GFDL/GFDL_atmos_cubed_sphere/blob/ab195d5026ca4c221b6cbb3888c8ae92d711f89a/driver/fvGFS/atmosphere.F90#L2199-L2200
        spfh_thresh = spfh.where(spfh > self.q_min, self.q_min)
        rTv = self.Rd * temp * (1. + self.z_vir*spfh_thresh)

        # layer mean pressure computed here:
        # https://github.com/NOAA-GFDL/GFDL_atmos_cubed_sphere/blob/ab195d5026ca4c221b6cbb3888c8ae92d711f89a/driver/fvGFS/atmosphere.F90#L2206-L2209
        # Numerator, our dpres = prsl in fortran code, as shown here: https://github.com/NOAA-GFDL/GFDL_atmos_cubed_sphere/blob/ab195d5026ca4c221b6cbb3888c8ae92d711f89a/driver/fvGFS/atmosphere.F90#L2122
        # Denominator is phii(k+1) - phii(k), which is equal to -g*delz,
        # see https://github.com/NOAA-GFDL/GFDL_atmos_cubed_sphere/blob/ab195d5026ca4c221b6cbb3888c8ae92d711f89a/driver/fvGFS/atmosphere.F90#L2125-L2131
        return dpres*rTv / (-self.g*delz)


    def calc_geopotential(self, hgtsfc: xr.DataArray, delz: xr.DataArray) -> xr.DataArray:
        """Compute geopotential field

        Note:
            If ``delz`` is computed by :meth:`calc_delz`, then a hydrostatic approximation is used. This may be inconsistent with the original simulation/dataset.

        Args:
            hgtsfc, delz (xr.DataArray): surface/orographic height, and vertical layer thickness

        Returns:
            geopotential (xr.DataArray): height in [m^2 / s^2] (i.e. height * gravity)
        """

        # a coordinate helper
        kp1_left = xr.DataArray(
            np.arange(len(self.pfull)),
            coords={self.level_name: self.pfull.values},
        )

        # Geopotential at the surface
        phi0 = self.g * hgtsfc
        phi0 = phi0.expand_dims({"kp1": [len(self.pfull)]})
        phi0 = phi0.reset_coords(drop=True)

        # Concatenate, cumulative sum from the ground to TOA
        dz = self.g*np.abs(delz)
        dz["kp1"] = kp1_left.sel({self.level_name: delz[self.level_name]})
        dz = dz.swap_dims({self.level_name: "kp1"}).drop_vars(self.level_name)
        dz = dz.reset_coords(drop=True)

        phii = xr.concat([dz,phi0], dim="kp1")

        phii = phii.sortby("kp1", ascending=False)
        phii = phii.cumsum("kp1")
        phii = phii.sortby("kp1", ascending=True)

        # At last, geopotential is interfacial value + .5 * layer thickness * gravity
        geopotential = phii - 0.5 * dz
        geopotential = geopotential.swap_dims({"kp1": self.level_name}).drop_vars("kp1")
        geopotential.attrs["units"] = "m**2 / s**2"
        geopotential.attrs["description"] = "Diagnosed using ufs2arco.Layers2Pressure.calc_geopotential"
        return geopotential

    def interp2pressure(
        self,
        xda: xr.DataArray,
        pstar: float,
        prsl: xr.DataArray,
        cds: Optional[xr.Dataset]=None,
    ):
        """Interpolate data on FV3 native vertical grid to pressure level (p*)

        Args:
            xda (xr.DataArray): field to be interpolated
            pstar (float): pressure level to be interpolated to, same units as prsl please
            prsl (xr.DataArray): layer mean pressure
            cds (xr.Dataset, optional): dataset with coefficients needed to do interpolation

        Returns:
            pda (xr.DataArray): field interpolated to pressure pstar
        """

        if cds is None:
            cds = self.get_interp_coefficients(pstar, prsl)

        xda_left = xda.where(cds["is_left"]).sum(self.level_name)
        xda_right = xda.where(cds["is_right"]).sum(self.level_name)

        result = xda_left + (xda_right - xda_left) * cds["factor"]

        mask = (cds["is_right"].sum(self.level_name) > 0) & (cds["is_left"].sum(self.level_name) > 0)
        return result.where(mask)


    def get_interp_coefficients(self, pstar: float, prsl: xr.DataArray) -> xr.Dataset:
        """Compute the coefficients needed to interpolate between pressure levels

        Args:
            pstar (float): pressure level to interpolate to in same units as prsl
            prsl (xr.DataArray): layer mean pressure

        Returns:
            cds (xr.Dataset): with coefficients needed to do interpolation
        """
        dlev = pstar - prsl
        dloglev = np.log(pstar) - np.log(prsl)

        # get coefficients representing pressure distance left and right of pstar
        is_left = dlev == dlev.where(dlev>=0).min(self.level_name)
        is_right = dlev == dlev.where(dlev<=0).max(self.level_name)

        # TODO: add extrapolation from level to ground here
        p_left = dloglev.where(is_left).sum(self.level_name)
        p_right = -dloglev.where(is_right).sum(self.level_name)

        denominator = p_left + p_right
        denominator = denominator.where(denominator > 1e-6, 1e-6)
        factor = p_left / denominator
        factor = factor.where(
            factor < 10.,
            10.
        ).where(
            factor > -10.,
            -10.,
        )
        return xr.Dataset({
            "is_left": is_left,
            "is_right": is_right,
            "p_left": p_left,
            "p_right": p_right,
            "denominator": denominator,
            "factor": factor,
        })

    def _dphalf_to_pfull(self, xda, name):
        xds = xda.to_dataset(name=name)
        xds = xds.assign_coords(self.pfull.coords)
        xds[self.level_name] = xds[self.level_name].swap_dims({self.level_name: self.interface_name})
        xds = xds.swap_dims({self.interface_name: self.level_name})
        xds = xds.drop_vars(self.interface_name)
        return xds[name]



from .base import Source
from .cloud_zarr import CloudZarrData
from .noaa_grib_forecast import NOAAGribForecastData

from .aws_aorc import AWSAORC
from .aws_gefs_archive import AWSGEFSArchive
from .aws_hrrr_archive import AWSHRRRArchive
from .gcs_era5_1degree import GCSERA5OneDegree
from .gcs_replay_atmosphere import GCSReplayAtmosphere
from .gfs_archive import GFSArchive


# writing something general is actually more work than
# just explicitly writing out the implemented data sources here
_recognized = {
    "aws_gefs_archive": "AWSGEFSArchive",
    "aws_hrrr_archive": "AWSHRRRArchive",
    "gcs_era5_1degree": "GCSERA5OneDegree",
    "gcs_replay_atmosphere": "GCSReplayAtmosphere",
    "gfs_archive": "GFSArchive",
    "aws_aorc": "AWSAORC",
}



import logging
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr

logger = logging.getLogger("ufs2arco")

class Source:
    """
    Base class for all datasets
    """

    # fill these out per subclass
    sample_dims = tuple()
    horizontal_dims = tuple()
    static_vars = tuple()
    available_variables = tuple()
    available_levels = tuple()

    @property
    def rename(self) -> dict:
        """
        Use this to map whatever the original source is to the ufs2arco standards...which need to be documented
        """
        return dict()

    @property
    def dynamic_vars(self) -> tuple:
        return tuple(x for x in self.variables if x not in self.static_vars)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def _level_sel_kwargs(self) -> dict:
        if self.use_nearest_levels:
            return {"method": "nearest"}
        else:
            return {}

    def __init__(
        self,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
        slices: Optional[dict] = None,
    ) -> None:
        """
        Initialize the Source object.

        Args:
            variables (list, tuple, optional): variables to grab
            levels (list, tuple, optional): vertical levels to grab
            use_nearest_levels (bool, optional): if True, all level selection with
                ``xarray.Dataset.sel(level=levels, method="nearest")``
            slices (dict, optional): either "sel" or "isel", with slice, passed to xarray
        """


        # check variable selection
        if variables is not None:
            unrecognized = []
            for v in variables:
                if v not in self.available_variables:
                    unrecognized.append(v)

            if len(unrecognized) > 0:
                raise NotImplementedError(f"{self.name}.__init__: the following variables are not recognized or not implemented: {unrecognized}")
            self.variables = variables
        else:
            self.variables = list(self.available_variables)

        # check level selection
        self.use_nearest_levels = use_nearest_levels
        if levels is not None:
            unrecognized = []
            for l in levels:
                if l not in self.available_levels:
                    unrecognized.append(l)

            if len(unrecognized) > 0 and not use_nearest_levels:
                msg = f"{self.name}.__init__: the following levels are not recognized or not implemented:\n{unrecognized}"
                msg += "\nSet option 'use_nearest_levels=True' if you want to use xarray.Dataset.sel(level=levels, method='nearest')"
                raise ValueError(msg)
        self.levels = levels

        # check slicing
        recognized = ("sel", "isel")
        if slices is not None:
            for key in slices.keys():
                if key not in recognized:
                    raise NotImplementedError(f"{self.name}.__init__: can't use {key} slice, only {recognized} are recognized so far...")
            for key, val in slices.items():
                if "level" in val and levels is not None:
                    logging.warning(f"{self.name}.__init__: you are using both the discrete 'levels' selection and specifying the slice-based selection xarray.Dataset.{key}({val}). Proceed at your own risk.")
            self.slices = slices
        else:
            self.slices = dict()

        logger.info(str(self))

    def __str__(self) -> str:
        title = f"Source: {self.name}"
        msg = f"\n{title}\n" + \
              "".join(["-" for _ in range(len(title))]) + "\n"
        attrslist = list(self.sample_dims) + ["variables", "levels", "use_nearest_levels", "slices"]
        for key in attrslist:
            msg += f"{key:<18s}: {getattr(self, key)}\n"
        if hasattr(self, "_accum_hrs"):
            key = "_accum_hrs"
            msg += f"{key:<18s}: {getattr(self, key)}\n"
        return msg

    def _open_static_vars(self, dims) -> bool:
        """Just do this once"""
        cond = True
        for key, val in dims.items():
            cond = cond and val == getattr(self, key)[0]
        return cond

    def add_full_extra_coords(self, xds: xr.Dataset) -> xr.Dataset:
        """
        An optional routine that builds extra coordinates if needed, see example in ensemble_forecast.py.
        This gets used for passive targets
        """
        return xds


    def apply_slices(self, xds: xr.Dataset) -> xr.Dataset:
        """Apply any slices, for now just data selection via "sel" or "isel"
        Note that this is the first transformation, so slicing options relate to
        the standard dimensions:

            (t0, fhr, member, level, latitude, longitude)

        Args:
            xds (xr.Dataset): The sample dataset

        Returns:
            xr.Dataset: The dataset after slices applied
        """

        if "sel" in self.slices.keys():
            for key, val in self.slices["sel"].items():
                xds = xds.sel({key: slice(*val)})

        if "isel" in self.slices.keys():
            for key, val in self.slices["isel"].items():
                xds = xds.isel({key: slice(*val)})
        return xds



"""
Rotate vectors, e.g. from model grid orientation to eastward/northward facing
"""

import numpy as np
import xarray as xr

def rotate_vectors(
    xds: xr.Dataset,
    vector_pairs: list[list[str, str]],
):

    for (u, v) in vector_pairs:
        xds[u], xds[v] = rotate_lambert_conical_vectors(xds[u], xds[v])

    return xds

def rotate_lambert_conical_vectors(u_grid, v_grid):
    """Note that this mirrors the calculations performed by NCEPLIBS-ip (often accessed via wgrib2)
    here: https://github.com/NOAA-EMC/NCEPLIBS-ip/blob/develop/src/ip_lambert_conf_grid_mod.F90

    See some verification of this here: https://gist.github.com/timothyas/4a0ea5a36b3015fd756a150be60d250f

    Args:
        u_grid, v_grid (xr.DataArray): the u, v vector pair, with data relative to model grid

    Returns:
        u_east, v_north (xr.DataArray): the u, v vector pair, with data pointing eastward and northward, respectively
    """
    uname = u_grid.name
    vname = v_grid.name
    not_here = []
    for key in [
        "GRIB_gridType",
        "GRIB_uvRelativeToGrid",
        "GRIB_latitudeOfFirstGridPointInDegrees",
        "GRIB_longitudeOfFirstGridPointInDegrees",
        "GRIB_Latin1InDegrees",
        "GRIB_Latin2InDegrees",
        "GRIB_LoVInDegrees",
    ]:
        try:
            assert key in u_grid.attrs
            assert key in v_grid.attrs
            assert u_grid.attrs[key] == v_grid.attrs[key]
        except:
            not_here.append(key)

    if len(not_here) > 0:
        raise AttributeError(f"rotate_lambert_conical_vectors: could not find the following GRIB attributes for the variable ({u_grid.name}, {v_grid.name}), or it's inconsistent between the two: {not_here}")

    assert u_grid.attrs["GRIB_gridType"] == "lambert", f"rotate_lambert_conical_vectors: only lambert conical rotation implemented, not true for ({u_grid.name}, {v_grid.name})"
    assert u_grid.attrs["GRIB_uvRelativeToGrid"] == 1, f"rotate_lambert_conical_vectors: only implemented for rotating from grid orientation to east/north, not true for ({u_grid.name}, {v_grid.name})"

    rlat1 = u_grid.attrs["GRIB_latitudeOfFirstGridPointInDegrees"]
    rlon1 = u_grid.attrs["GRIB_longitudeOfFirstGridPointInDegrees"]
    rlati1 = u_grid.attrs["GRIB_Latin1InDegrees"]
    rlati2 = u_grid.attrs["GRIB_Latin2InDegrees"]
    orient = u_grid.attrs["GRIB_LoVInDegrees"]


    # now, do the actual work
    eps = np.finfo(float).eps

    if np.abs(rlati1 - rlati2) < eps:
        angle = np.sin(np.deg2rad(rlati1))

    else:
        angle = np.log(
            np.cos(np.deg2rad(rlati1))/np.cos(np.deg2rad(rlati2))
        ) / np.log(
            np.tan(np.deg2rad((90-rlati1)/2))/np.tan(np.deg2rad((90-rlati2)/2))
        )

    dlon = (u_grid["longitude"] - orient + 180 + 3600 % 360) - 180

    crot = np.cos(angle*np.deg2rad(dlon)).astype(u_grid.dtype)
    srot = np.sin(angle*np.deg2rad(dlon)).astype(u_grid.dtype)

    with xr.set_options(keep_attrs=True):
        u_east  = u_grid * crot + v_grid * srot
        v_north = v_grid * crot - u_grid * srot

    # manage coords
    u_east = u_east.rename(uname)
    v_north = v_north.rename(vname)
    u_east.attrs["GRIB_uvRelativeToGrid"] = 0
    v_north.attrs["GRIB_uvRelativeToGrid"] = 0
    u_east.attrs["GRIB_latitudeOfSouthernPoleInDegrees"] = -90.0
    v_north.attrs["GRIB_latitudeOfSouthernPoleInDegrees"] = -90.0

    u_east.attrs["rotation"] = "rotated to point eastward via ufs2arco.transforms.rotate_lambert_conical_vectors"
    v_north.attrs["rotation"] = "rotated to point northward via ufs2arco.transforms.rotate_lambert_conical_vectors"
    return u_east, v_north



import logging
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr

from ufs2arco.sources import Source, NOAAGribForecastData

logger = logging.getLogger("ufs2arco")

class GFSArchive(NOAAGribForecastData, Source):
    """
    Access 1/4 degree archives of NOAA's Global Forecast System (GFS) via:
        * if before 2021: UCAR Research Data Archive (RDA)
            * https://rda.ucar.edu/datasets/d084001
            * https://rda.ucar.edu/datasets/d084003
        * after 2021: AWS at https://registry.opendata.aws/noaa-gfs-bdp-pds/
    """

    sample_dims = ("t0", "fhr")
    horizontal_dims = ("latitude", "longitude")
    file_suffixes = ("", "b")
    static_vars = ("lsm", "orog")

    @property
    def available_levels(self) -> tuple:
        return (
            1, 2, 3, 5, 7, 10, 15,
            20, 30, 40, 50, 70,
            100, 125, 150, 175, 200, 225, 250, 275,
            300, 325, 350, 375, 400, 425, 450, 475,
            500, 525, 550, 575, 600, 625, 650, 675,
            700, 725, 750, 775, 800, 825, 850, 875,
            900, 925, 950, 975, 1000,
        )

    @property
    def rename(self) -> dict:
        return {
            "time": "t0",
            "step": "lead_time",
            "isobaricInhPa": "level",
        }

    def __init__(
        self,
        t0: dict,
        fhr: dict,
        variables: Optional[list | tuple] = None,
        levels: Optional[list | tuple] = None,
        use_nearest_levels: Optional[bool] = False,
        slices: Optional[dict] = None,
        accum_hrs: Optional[dict] = None,
    ) -> None:
        """
        Args:
            t0 (dict): Dictionary with start and end times for initial conditions, and e.g. "freq=6h". All options get passed to ``pandas.date_range``.
            fhr (dict): Dictionary with 'start', 'end', and 'step' forecast hours.
            variables (list, tuple, optional): variables to grab
            levels (list, tuple, optional): vertical levels to grab
            use_nearest_levels (bool, optional): if True, all level selection with
                ``xarray.Dataset.sel(level=levels, method="nearest")``
            slices (dict, optional): either "sel" or "isel", with slice, passed to xarray
            accum_hrs (dict, optional): period in hours over which to accumulate accumulated variables, e.g.
                {"accum_tp": 1}
                would be the same as passing filter_by_keys={"stepRange": "5-6"}
                when reading fhr 6 data using xarray and cfgrib
        """
        self.t0 = pd.date_range(**t0)
        self.fhr = np.arange(fhr["start"], fhr["end"] + 1, fhr["step"])
        super().__init__(
            variables=variables,
            levels=levels,
            use_nearest_levels=use_nearest_levels,
            slices=slices,
            accum_hrs=accum_hrs,
        )


        # make sure the default variable set "just works"
        # by pulling out the variable names only available at forecast time
        if tuple(self.variables) == self.available_variables:
            varlist = []
            for key, meta in self._varmeta.items():
                if not self._varmeta[key]["forecast_only"]:
                    varlist.append(key)
            self.variables = varlist

        # for GFS, plenty of variables only exist in the forecast, not analysis grib files
        # make sure the user doesn't ask for these before we get started
        if any(self.fhr == 0):
            requested_vars_not_in_analysis = []
            for varname in self.variables:
                if self._varmeta[varname]["forecast_only"]:
                    requested_vars_not_in_analysis.append(varname)
            if len(requested_vars_not_in_analysis) > 0:
                msg = f"{self.name}.__init__: the following requested variables only exist in forecast timesteps"
                msg += f"\n\n{requested_vars_not_in_analysis}\n\n"
                msg += "these should be requested separately with only fhr > 0 (i.e., fhr start >0 in your yaml)"
                raise Exception(msg)

    def _build_path(
        self,
        t0: pd.Timestamp,
        fhr: int,
        file_suffix: str,
    ) -> str:
        """
        Build the file path to a GFS GRIB file on RDA or AWS.

        Args:
            t0 (pd.Timestamp): The initial condition timestamp.
            fhr (int): The forecast hour.
            file_suffix (str): For GFS, either '' or 'b'.

        Returns:
            str: The constructed file path.
        """
        if t0 < pd.Timestamp("2021-01-01T00"):

            bucket = f"https://data.rda.ucar.edu/d084001" if file_suffix == "" else \
                    f"https://data.rda.ucar.edu/d084003"
            outer = f"{t0.year:04d}/{t0.year:04d}{t0.month:02d}{t0.day:02d}"
            fname = f"gfs.0p25{file_suffix}.{t0.year:04d}{t0.month:02d}{t0.day:02d}{t0.hour:02d}.f{fhr:03d}.grib2"


        else:
            bucket = "s3://noaa-gfs-bdp-pds"
            outer = f"gfs.{t0.year:04d}{t0.month:02d}{t0.day:02d}/{t0.hour:02d}"
            fname = f"gfs.t{t0.hour:02d}z.pgrb2{file_suffix}.0p25.f{fhr:03d}"

            if t0 > pd.Timestamp("2021-03-22T06"):
                fname = "atmos/" + fname

        fullpath = f"filecache::{bucket}/{outer}/{fname}"
        logger.debug(f"{self.name}._build_path: reading {fullpath}")
        return fullpath



"""We can't use a dask cluster, because it cannot serialize the tasks of opening multiple
datasets with an io buffered reader object, orsomething.

So, this is easy enough to just submit separate slurm jobs that work on their own job ID.
"""

import os
import subprocess

from ufs2arco import FV3Dataset, Timer
from replay_mover import ReplayMover1Degree


def submit_slurm_mover(job_id, mover):

    the_code = \
        f"from replay_mover import ReplayMover1Degree\n"+\
        f"mover = ReplayMover1Degree(\n"+\
        f"    n_jobs={mover.n_jobs},\n"+\
        f"    config_filename='{mover.config_filename}',\n"+\
        f"    storage_options={mover.storage_options},\n"+\
        f"    main_cache_path='{mover.main_cache_path}',\n"+\
        f"    component='{mover.component}',\n"+\
        f")\n"+\
        f"mover.run({job_id})"

    slurm_dir = "slurm/replay-1.00-degree"
    txt = "#!/bin/bash\n\n" +\
        f"#SBATCH -J r1d{job_id:03d}\n"+\
        f"#SBATCH -o {slurm_dir}/{job_id:03d}.%j.out\n"+\
        f"#SBATCH -e {slurm_dir}/{job_id:03d}.%j.err\n"+\
        f"#SBATCH --nodes=1\n"+\
        f"#SBATCH --ntasks=1\n"+\
        f"#SBATCH --cpus-per-task=30\n"+\
        f"#SBATCH --partition=compute\n"+\
        f"#SBATCH -t 120:00:00\n\n"+\
        f"source /contrib/Tim.Smith/miniconda3/etc/profile.d/conda.sh\n"+\
        f"conda activate ufs2arco\n"+\
        f'python -c "{the_code}"'

    script_dir = "job-scripts"
    fname = f"{script_dir}/submit_1mover{job_id:03d}.sh"

    for this_dir in [slurm_dir, script_dir]:
        if not os.path.isdir(this_dir):
            os.makedirs(this_dir)

    with open(fname, "w") as f:
        f.write(txt)

    subprocess.run(f"sbatch {fname}", shell=True)


if __name__ == "__main__":

    walltime = Timer()
    localtime = Timer()

    walltime.start("Initializing job")

    mover = ReplayMover1Degree(
        n_jobs=15,
        config_filename="config-1.00-degree.yaml",
        storage_options={"token": "/contrib/Tim.Smith/.gcs/replay-service-account.json"},
        main_cache_path="/lustre/Tim.Smith/tmp-replay/1.00-degree",
        component="fv3",
    )

    localtime.start("Make and Store Container Dataset")
    mover.store_container()
    localtime.stop()

    localtime.start("Run slurm jobs")
    for job_id in range(mover.n_jobs):
        submit_slurm_mover(job_id, mover)
    localtime.stop()

    walltime.stop("Walltime Time")



import os
import sys
import logging
import yaml

import numpy as np
import xarray as xr
import pytest

from ufs2arco.driver import Driver
from ufs2arco.multidriver import MultiDriver
from ufs2arco.log import SimpleFormatter

logger = logging.getLogger("integration-test")
_local_path = os.path.dirname(__file__)
_sources = ["replay", "gfs", "gefs", "hrrr", "era5"]
_targets = ["base", "anemoi"]

def setup_test_log():
    logger.setLevel(logging.INFO)

    # Create a console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)

    # Define a simple formatter
    formatter = SimpleFormatter(fmt="[%(relativeCreated)d s] [%(levelname)-7s] %(message)s")
    console_handler.setFormatter(formatter)

    # Attach the handler to the test logger
    logger.addHandler(console_handler)

    # Ensure the test logger does not propagate messages to the root logger
    logger.propagate = False


def create_regrid_target_grid(filepath):
    import xesmf
    ds = xesmf.util.grid_global(1, 1, cf=True, lon1=360)
    ds = ds.drop_vars("latitude_longitude")

    # GFS goes north -> south
    ds = ds.sortby("lat", ascending=False)

    ds.to_netcdf(filepath)

def run_test(source, target):
    logger.info(f"Starting Test: {source} {target}")

    fname = os.path.join(
        _local_path,
        f"{source}.{target}.yaml",
    )
    with open(fname, "r") as f:
        config = yaml.safe_load(f)

    # setup directory for this test
    test_dir = os.path.join(
        _local_path,
        source,
        target,
    )
    if not os.path.isdir(test_dir):
        os.makedirs(test_dir)
        logger.info(f"Creating {test_dir}")

    # prepend directories and filenames with test directory path
    for key in config["directories"].keys():
        val = config["directories"][key]
        config["directories"][key] = os.path.join(test_dir, val)

    if "horizontal_regrid" in config.get("transforms", {}):
        filepath = config["transforms"]["horizontal_regrid"]["target_grid_path"]
        filepath = os.path.join(test_dir, filepath)
        config["transforms"]["horizontal_regrid"]["target_grid_path"] = filepath
        create_regrid_target_grid(filepath)

    # write a specific config
    config_filename = os.path.join(test_dir, "config.yaml")
    with open(config_filename, "w") as f:
        yaml.dump(config, stream=f)

    # run driver
    if "multisource" in config.keys():
        driver = MultiDriver(config_filename)
    else:
        driver = Driver(config_filename)
    driver.run(overwrite=True)

    # read & print last line of log
    logfile = os.path.join(
        config["directories"]["logs"],
        "log.serial.out",
    )
    with open(logfile, "rb") as f:
        f.seek(-2, 2)  # Move to the second-to-last byte of the file
        while f.read(1) != b"\n":  # Move backward until finding a newline
            f.seek(-2, 1)
        last_line = f.readline().decode()  # Read the last line
    logger.info(last_line)

    # now run the tests
    _test_static_vars(source, target, config["directories"]["zarr"])

    logger.info(f" ... Test Passed")

def _test_static_vars(source, target, store):
    ds = xr.open_zarr(store, decode_timedelta=True)

    lsm = {
        "gefs": "lsm",
        "gfs": "lsm",
        "hrrr": "lsm",
        "replay": "land_static",
        "era5": "round_land_sea_mask",
    }[source]

    orog = {
        "gefs": "orog",
        "gfs": "orog",
        "hrrr": "orog",
        "replay": "hgtsfc_static",
        "era5": "orography",
    }[source]

    # test land sea mask
    for varname in [lsm, orog]:
        if target == "anemoi":
            idx = ds.attrs["variables"].index(varname)
            xda = ds["data"].sel(variable=idx)
        else:
            xda = ds[varname]

        assert np.all(~np.isnan(xda.values)), f"Found NaNs in {source} {target} {varname}"
        if varname == lsm:
            np.testing.assert_almost_equal(
                xda.max().values,
                1,
                err_msg=f"Found max != 1 in {source} {target} {varname}",
            )
            np.testing.assert_almost_equal(
                xda.min().values,
                0,
                err_msg=f"Found min != 1 in {source} {target} {varname}",
            )
        elif varname == orog:

            # orography mean should be like ~300
            # make sure it's not geopotential at surface, which is 9.81x that val
            # so taking log should be OK as a test?
            # this is just to make sure it's meaningful...
            val = np.floor(np.log10(xda.mean().values))
            np.testing.assert_equal(
                val,
                2,
                err_msg=f"Found floor( log10( orography.mean() )) = {val} != 2 for {source} {target} {varname}",
            )



@pytest.fixture(scope="module", autouse=True)
def setup_logging():
    setup_test_log()

@pytest.mark.dependency()
@pytest.mark.parametrize("target", _targets)
@pytest.mark.parametrize("source", _sources)
def test_this_combo(source, target):
    run_test(source, target)

@pytest.mark.dependency(
    depends=[
        f"test_this_combo[{source}-{target}]"
        for source in _sources
        for target in _targets
    ]
)
@pytest.mark.parametrize("source", _sources)
def test_flattened_base_equals_anemoi(source):
    ds = xr.open_zarr(os.path.join(_local_path, source, "base", "dataset.zarr"), decode_timedelta=True)
    if "pressure" in ds.dims:
        ds = ds.rename({"pressure": "level"})
    if "t0" in ds.dims:
        ds = ds.rename({"t0": "time"})
        ds = ds.sel(fhr=0, drop=True)
    ads = xr.open_zarr(os.path.join(_local_path, source, "anemoi", "dataset.zarr"), decode_timedelta=True)

    for key in ds.data_vars:

        if "level" in ds[key].dims:
            for level in ds[key].level.values:
                ilevel = int(level) if level==int(level) else level
                idx = ads.attrs["variables"].index(f"{key}_{ilevel}")
                for itime in ads.time.values:
                    np.testing.assert_allclose(
                        ds[key].isel(time=itime).sel(level=level).squeeze().values.flatten(),
                        ads["data"].sel(variable=idx, time=itime).squeeze().values.flatten(),
                    )
        else:
            idx = ads.attrs["variables"].index(key)
            for itime in ads.time.values:
                for imember in ads.ensemble.values:
                    xda = ds[key].isel(time=itime) if "time" in ds[key].dims else ds[key]
                    xda = xda.isel(member=imember) if "member" in ds[key].dims else xda
                    np.testing.assert_allclose(
                        xda.squeeze().values.flatten(),
                        ads["data"].sel(variable=idx, time=itime, ensemble=imember).squeeze().values.flatten(),
                    )



import os
import subprocess

from ufs2arco import FV3Dataset, Timer
from replay_mover import ReplayMoverQuarterDegree


def submit_slurm_filler(job_id, mover):

    the_code = \
        f"from replay_mover import ReplayMoverQuarterDegree\n"+\
        f"from replay_filler import ReplayFiller\n"+\
        f"mover = ReplayMoverQuarterDegree(\n"+\
        f"    n_jobs={mover.n_jobs},\n"+\
        f"    config_filename='{mover.config_filename}',\n"+\
        f"    storage_options={mover.storage_options},\n"+\
        f"    main_cache_path='{mover.main_cache_path}',\n"+\
        f")\n"+\
        f"filler = ReplayFiller(mover=mover)\n"+\
        f"filler.run({job_id})"

    slurm_dir = "slurm/replay-fill-0.25-degree"
    txt = "#!/bin/bash\n\n" +\
        f"#SBATCH -J rfqd{job_id:03d}\n"+\
        f"#SBATCH -o {slurm_dir}/{job_id:03d}.%j.out\n"+\
        f"#SBATCH -e {slurm_dir}/{job_id:03d}.%j.err\n"+\
        f"#SBATCH --nodes=1\n"+\
        f"#SBATCH --ntasks=1\n"+\
        f"#SBATCH --cpus-per-task=30\n"+\
        f"#SBATCH --partition=compute\n"+\
        f"#SBATCH -t 120:00:00\n\n"+\
        f"source /contrib/Tim.Smith/miniconda3/etc/profile.d/conda.sh\n"+\
        f"conda activate ufs2arco\n"+\
        f'python -c "{the_code}"'

    script_dir = "job-scripts"
    fname = f"{script_dir}/submit_qfiller{job_id:03d}.sh"

    for this_dir in [slurm_dir, script_dir]:
        if not os.path.isdir(this_dir):
            os.makedirs(this_dir)

    with open(fname, "w") as f:
        f.write(txt)

    subprocess.run(f"sbatch {fname}", shell=True)


if __name__ == "__main__":

    walltime = Timer()
    localtime = Timer()

    walltime.start("Initializing job")

    # Recreated from what we used to put up quarter degree in the first place
    mover = ReplayMoverQuarterDegree(
        n_jobs=1,
        config_filename="config-0.25-degree.yaml",
        storage_options={"token": "/contrib/Tim.Smith/.gcs/replay-service-account.json"},
        main_cache_path="/lustre/Tim.Smith/tmp-replay/0.25-degree",
    )

    localtime.start("Run slurm jobs")
    for job_id in range(mover.n_jobs):
        submit_slurm_filler(job_id, mover)
    localtime.stop()

    walltime.stop("Walltime Time")



import logging

import numpy as np
import xarray as xr

logger = logging.getLogger("ufs2arco")

def get_available_mappings() -> dict:
    return {
        "log": _log,
        "round": _round,
    }

def apply_mappings(
    xds: xr.Dataset,
    mappings: dict,
) -> xr.Dataset:
    """
    Apply any of the recognized mappings listed in :func:`get_available_mappings`

    Args:
        xds (xr.Dataset): with pre-mapped data
        mappings (dict): with a dictionary with keys as function names, and list of variables to apply it to, e.g.
            {"log": ["spfh", "spfh2m"]}

    Returns:
        xds (xr.Dataset): with mapped data, and originally mapped fields removed
    """

    function_mappings = get_available_mappings()

    for mapname, varlist in mappings.items():
        varlist = [varlist] if isinstance(varlist, str) else varlist
        func = function_mappings[mapname]
        for varname in varlist:
            if varname in xds:
                new_name = f"{mapname}_{varname}"
                xds[new_name] = func(xds[varname])
                long_name = xds[varname].attrs.get("long_name", varname)
                xds[new_name].attrs["long_name"] = f"{mapname} of {long_name}"
                xds = xds.drop_vars(varname)
    return xds

def _log(xda):
    """Performs a protected log of the provided array"""
    cond = xda > 0
    lda = xr.where(
        cond,
        np.log(xda.where(cond)),
        0.,
    )

    # unsure if this is truly what we should do... but whatever
    lda.attrs = xda.attrs.copy()
    lda.attrs["units"] = ""
    return lda

def _round(xda):
    """apply numpy.round"""
    with xr.set_options(keep_attrs=True):
        return np.round(xda)




import logging

import xarray as xr

from ufs2arco.transforms.horizontal_regrid import horizontal_regrid
from ufs2arco.transforms.mappings import get_available_mappings, apply_mappings
from ufs2arco.transforms.rotate_vectors import rotate_vectors
from ufs2arco.transforms.vertical_regrid import fv_vertical_regrid

logger = logging.getLogger("ufs2arco")

class Transformer:

    @property
    def implemented(self) -> tuple:
        return (
            "multiply",
            "divide",
            "fv_vertical_regrid",
            "horizontal_regrid",
            "mappings",
            "rotate_vectors",
        )

    def __init__(self, options):

        names = list(options.keys())

        # first check regrid, mappings, etc
        unrecognized = []
        for name in names:
            if name not in self.implemented:
                unrecognized.append(name)

        if len(unrecognized) > 0:
            raise NotImplementedError(f"Transformer.__init__: the following transformations are not recognized or not implemented: {unrecognized}")

        # now check for mappings
        unrecognized = []
        if "mappings" in names:
            available = list(get_available_mappings().keys())
            for mapname in options["mappings"]:
                if mapname not in available:
                    unrecognized.append(mapname)

            if len(unrecognized) > 0:
                raise NotImplementedError(f"Transformer.__init__: the following mappings are not recognized or not implemented: {unrecognized}")

        # if we want to do horizontal regridding, check if xesmf is installed
        if "horizontal_regrid" in names:
            try:
                import xesmf
            except ImportError:
                raise ImportError(f"Transformer.__init__: Could not 'import xesmf', but this is needed for 'horizontal_grid' transformations. Install xesmf with\n'conda install -c conda-forge xesmf'")

        self.names = names
        self.options = options
        logger.info(str(self))

    def __str__(self) -> str:
        title = f"Transformations"
        msg = f"\n{title}\n" + \
              "".join(["-" for _ in range(len(title))]) + "\n"
        optstr = "\n    ".join([f"{key:<14s}: {val}" for key, val in self.options.items()])
        msg += f"options\n    {optstr}\n"
        return msg

    def __call__(self, xds: xr.Dataset):
        """
        Process the dataset, performing any of the desired transformations

        Args:
            xds (xr.Dataset): with the data

        Returns:
            xds (xr.Dataset): the processed version
        """

        if "multiply" in self.names:
            xds = multiply(xds, self.options["multiply"])

        if "divide" in self.names:
            xds = divide(xds, self.options["divide"])

        if "rotate_vectors" in self.names:
            xds = rotate_vectors(xds, **self.options["rotate_vectors"])

        if "fv_vertical_regrid" in self.names:
            xds = fv_vertical_regrid(xds, **self.options["fv_vertical_regrid"])

        if "horizontal_regrid" in self.names:
            xds = horizontal_regrid(xds, **self.options["horizontal_regrid"])

        if "mappings" in self.names:
            xds = apply_mappings(xds, self.options["mappings"])

        return xds

def multiply(xds, config):
    """
    Multiply selected variables by a scalar, or if it can be provided in a yaml, an array of
    an appropriately broadcastable size

    Note that for now, attrs are not preserved in this process, for hopefully obvious reasons

    Args:
        xds (xr.Dataset): the dataset from source
        config (dict): with pattern
            {varname: scalar_value_to_multiply_by}

    Returns:
        xds (xr.Dataset): with new dataset
    """
    for varname, scalar in config.items():
        if varname in xds:
            xds[varname] = xds[varname] * scalar
    return xds

def divide(xds, config):
    """
    Divide selected variables by a scalar, or if it can be provided in a yaml, an array of
    an appropriately broadcastable size

    Note that for now, attrs are not preserved in this process, for hopefully obvious reasons

    Args:
        xds (xr.Dataset): the dataset from source
        config (dict): with pattern
            {varname: scalar_value_to_divide_by}

    Returns:
        xds (xr.Dataset): with new dataset
    """
    for varname, scalar in config.items():
        if varname in xds:
            xds[varname] = xds[varname] / scalar
    return xds



import os
import logging
from typing import Optional

import numpy as np
import dask
import pandas as pd
import xarray as xr

import cf_xarray as cfxr

try:
    import xesmf

    _has_xesmf = True
except ImportError:
    _has_xesmf = False

from ufs2arco.regrid.gaussian_grid import gaussian_latitudes

logger = logging.getLogger("ufs2arco")

def horizontal_regrid(
    xds: xr.Dataset,
    target_grid_path: str,
    regridder_kwargs: dict,
    open_target_kwargs: Optional[dict] = None,
    source_is_on_gaussian_grid: bool = False,
):

    kw = {} if open_target_kwargs is None else open_target_kwargs
    ds_out = xr.open_dataset(os.path.expandvars(target_grid_path), **kw)

    # required for xesmf
    rename = {"longitude": "lon", "latitude": "lat"}
    for key, val in rename.items():
        if key in xds:
            xds = xds.rename({key: val})

    # for the first time, we have to compute the regridder weights no matter what
    # so, figure out if we have the file or not
    kw = regridder_kwargs.copy()
    filename = kw.get(
        "filename",
        f"{kw['method']}_{len(xds.lat)}x{len(xds.lon)}_{len(ds_out.lat)}x{len(ds_out.lon)}.nc",
    )
    filename = os.path.expandvars(filename)
    kw["filename"] = filename
    if os.path.isfile(filename):
        kw["reuse_weights"] = kw.get("reuse_weights", True)
    else:
        kw["reuse_weights"] = False
        logger.info(f"ufs2arco.transforms.horizontal_regrid: couldn't find xesmf weights filename {filename}, they'll be computed now.")

    # check if bounds are there
    if "lat_b" not in xds and "lon_b" not in xds:
        if not os.path.isfile(filename):
            logger.info(f"ufs2arco.transforms.horizontal_regrid: did not find 'lat_b' or 'lon_b' in source, computing bounds.")
        xds = get_bounds(xds, is_gaussian=source_is_on_gaussian_grid)
        if not os.path.isfile(filename):
            logger.info(f"ufs2arco.transforms.horizontal_regrid: computed bounds are\nLongitude\n{xds.lon_b}\nLatitude\n{xds.lat_b}")

    # check input_array for c_contiguous or not
    xds = maybe_make_dataset_c_contiguous(xds)

    # do the work
    regridder = xesmf.Regridder(
        ds_in=xds,
        ds_out=ds_out,
        **kw,
    )
    ds_out = regridder(xds, keep_attrs=True)

    # ds_out has the lat/lon boundaries from input dataset
    # remove these because it doesn't make sense anymore
    ds_out = ds_out.drop_vars(["lat_b", "lon_b"])

    # this one is a weird one, created by xesmf's global grid util creator
    if "latitude_longitude" in ds_out:
        if np.isnan(ds_out["latitude_longitude"]):
            ds_out = ds_out.drop_vars("latitude_longitude")
    ds_out = ds_out.rename({"lon": "longitude", "lat": "latitude",})
    return ds_out

def get_bounds(xds, is_gaussian=False):
    """
    Use cf_xarray to get the bounds of a rectilinear grid
    """
    xds = xds.cf.add_bounds(["lat", "lon"])

    for key in ["lat", "lon"]:
        corners = cfxr.bounds_to_vertices(
            bounds=xds[f"{key}_bounds"],
            bounds_dim="bounds",
            order=None,
        )
        xds = xds.assign_coords({f"{key}_b": corners})
        xds = xds.drop_vars(f"{key}_bounds")

    if is_gaussian:
        xds = xds.drop_vars("lat_b")
        _, lat_b = gaussian_latitudes(len(xds.lat)//2)
        lat_b = np.concatenate([lat_b[:,0], [lat_b[-1,-1]]])
        if xds["lat"][0] > 0:
            lat_b = lat_b[::-1]
        xds["lat_b"] = xr.DataArray(
            lat_b,
            dims="lat_vertices",
        )
        xds = xds.set_coords("lat_b")
    return xds


def create_output_dataset(lat, lon, is_gaussian):
    xds = xr.Dataset({
        "lat": lat,
        "lon": lon,
    })
    return get_bounds(xds, is_gaussian)

def maybe_make_dataset_c_contiguous(xds: xr.Dataset) -> xr.Dataset:
    """xesmf expects arrays to be C_CONTIGUOUS, conforming to python data standards.
    However, GCM data is often F_CONTIGUOUS, since most numerical cores are in fortran.
    Deal with that here...

    Args:
        xds (xr.Dataset): with the input data, pre regridding

    Returns:
        xds (xr.Dataset): with original data if it's already C_CONTIGUOUS, or
            with data modified to be C_CONTIGUOUS
    """
    modified = False
    new_vars = {}

    for name, xda in xds.data_vars.items():
        data = xda.data
        if hasattr(data, "flags") and not data.flags["C_CONTIGUOUS"]:
            data = np.ascontiguousarray(data)
            modified = True
            new_vars[name] = xr.DataArray(data, dims=xda.dims, coords=xda.coords, attrs=xda.attrs)
        else:
            new_vars[name] = xda

    if modified:
        return xr.Dataset(new_vars, coords=xds.coords, attrs=xds.attrs)
    else:
        return xds
